# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from asyncio import gather, iscoroutine, sleep, ensure_future
from dataclasses import dataclass
from typing import Dict, TYPE_CHECKING, Union, Optional

from .command_types import AppCommandOptionType
from .interaction_base import InteractionType, CallbackType
from ..app.select_menu import SelectOption
from ..guild.member import GuildMember
from ..message.context import MessageContext
from ..message.message import Message
from ..message.user_message import UserMessage
from ..user import User
from ...exceptions import InteractionDoesNotExist, UseFollowup, \
    InteractionAlreadyAcknowledged, NotFoundError, InteractionTimedOut
from ...utils import APIObject, convert
from ...utils.convert_message import convert_message
from ...utils.snowflake import Snowflake
from ...utils.types import MISSING

if TYPE_CHECKING:
    from .interaction_flags import InteractionFlags
    from ...utils.convert_message import MessageConvertable
    from .command import AppCommandInteractionDataOption
    from ..guild.channel import Channel
    from ..guild.role import Role
    from ...utils import APINullable


@dataclass
class ResolvedData(APIObject):
    """Represents a Discord Resolved Data structure

    Attributes
    ----------
    users: APINullable[Dict[:class:`~pincer.utils.snowflake.Snowflake`, :class:`~pincer.objects.user.user.User`]]
        Map of Snowflakes to user objects
    members: APINullable[Dict[:class:`~pincer.utils.snowflake.Snowflake`, :class:`~pincer.objects.guild.member.GuildMember`]]
        Map of Snowflakes to partial member objects
    roles: APINullable[Dict[:class:`~pincer.utils.snowflake.Snowflake`, :class:`~pincer.objects.guild.role.Role`]]
        Map of Snowflakes to role objects
    channels: APINullable[Dict[:class:`~pincer.utils.snowflake.Snowflake`, :class:`~pincer.objects.guild.channel.Channel`]]
        Map of Snowflakes to partial channel objects
    messages: APINullable[Dict[:class:`~pincer.utils.snowflake.Snowflake`, :class:`~pincer.objects.message.user_message.UserMessage`]]
        Map of Snowflakes to partial message objects
    """
    # noqa: E501
    users: APINullable[Dict[Snowflake, User]] = MISSING
    members: APINullable[Dict[Snowflake, GuildMember]] = MISSING
    roles: APINullable[Dict[Snowflake, Role]] = MISSING
    channels: APINullable[Dict[Snowflake, Channel]] = MISSING
    messages: APINullable[Dict[Snowflake, UserMessage]] = MISSING


@dataclass
class InteractionData(APIObject):
    """Represents a Discord Interaction Data structure

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The `ID` of the invoked command
    name: :class:`str`
        The `name` of the invoked command
    type: :class:`int`
        The `type` of the invoked command
    resolved: APINullable[:class:`~pincer.objects.app.interactions.ResolvedData`]
        Converted users + roles + channels
    options: APINullable[:class:`~pincer.objects.app.command.AppCommandInteractionDataOption`]
        The params + values from the user
    custom_id: APINullable[:class:`str`]
        The `custom_id` of the component
    component_type: APINullable[:class:`int`]
        The type of the component
    values: APINullable[:class:`~pincer.objects.app.select_menu.SelectOption`]
        The values the user selected
    target_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the user or message targeted by a user or message command
    """
    # noqa: E501
    id: Snowflake
    name: str
    type: int

    resolved: APINullable[ResolvedData] = MISSING
    options: APINullable[AppCommandInteractionDataOption] = MISSING
    custom_id: APINullable[str] = MISSING
    component_type: APINullable[int] = MISSING
    values: APINullable[SelectOption] = MISSING
    target_id: APINullable[Snowflake] = MISSING


@dataclass
class Interaction(APIObject):
    """Represents a Discord Interaction object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the interaction
    application_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the application this interaction is for
    type: :class:`~pincer.objects.app.interaction_base.InteractionType`
        The type of interaction
    token: :class:`str`
        A continuation token for responding to the interaction
    version: :class:`int`
        Read-only property, always ``1``
    data: APINullable[:class:`~pincer.objects.app.interactions.InteractionData`]
        The command data payload
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The guild it was sent from
    channel_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The channel it was sent from
    member: APINullable[:class:`~pincer.objects.guild.member.GuildMember`]
        Guild member data for the invoking user, including permissions
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        User object for the invoking user, if invoked in a DM
    message: APINullable[:class:`~pincer.objects.message.user_message.UserMessage`]
        For components, the message they were attached to
    """
    # noqa: E501
    id: Snowflake
    application_id: Snowflake
    type: InteractionType
    token: str

    version: int = 1
    data: APINullable[InteractionData] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    channel_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING
    user: APINullable[User] = MISSING
    message: APINullable[UserMessage] = MISSING
    has_replied: bool = False
    has_acknowledged: bool = False

    def __post_init__(self):
        self.id = convert(self.id, Snowflake.from_string)
        self.application_id = convert(
            self.application_id, Snowflake.from_string
        )
        self.type = convert(self.type, InteractionType)
        self.data = convert(
            self.data,
            InteractionData.from_dict,
            InteractionData
        )
        self.guild_id = convert(self.guild_id, Snowflake.from_string)
        self.channel_id = convert(self.channel_id, Snowflake.from_string)

        self.member = convert(
            self.member,
            GuildMember.from_dict,
            GuildMember,
            client=self._client
        )

        self.user = convert(
            self.user,
            User.from_dict,
            User,
            client=self._client
        )

        self.message = convert(
            self.message,
            UserMessage.from_dict,
            UserMessage,
            client=self._client
        )

        self._convert_functions = {
            AppCommandOptionType.SUB_COMMAND: None,
            AppCommandOptionType.SUB_COMMAND_GROUP: None,

            AppCommandOptionType.STRING: str,
            AppCommandOptionType.INTEGER: int,
            AppCommandOptionType.BOOLEAN: bool,
            AppCommandOptionType.NUMBER: float,

            AppCommandOptionType.USER: lambda value:
            self._client.get_user(
                convert(value, Snowflake.from_string)
            ),
            AppCommandOptionType.CHANNEL: lambda value:
            self._client.get_channel(
                convert(value, Snowflake.from_string)
            ),
            AppCommandOptionType.ROLE: lambda value:
            self._client.get_role(
                convert(self.guild_id, Snowflake.from_string),
                convert(value, Snowflake.from_string)
            ),
            AppCommandOptionType.MENTIONABLE: None
        }

    async def build(self):
        """|coro|

        Sets the parameters in the interaction that need information
        from the discord API.
        """
        if not self.data.options:
            return

        await gather(
            *map(self.convert, self.data.options)
        )

    async def convert(self, option: AppCommandInteractionDataOption):
        """|coro|

        Sets an AppCommandInteractionDataOption value parameter to
        the payload type
        """
        converter = self._convert_functions.get(option.type)

        if not converter:
            raise NotImplementedError(
                f"Handling for AppCommandOptionType {option.type} is not "
                "implemented"
            )

        res = converter(option.value)

        option.value = (await res) if iscoroutine(res) else res

    def convert_to_message_context(self, command):
        return MessageContext(
            self.member or self.user,
            command,
            self,
            self.guild_id,
            self.channel_id
        )

    async def response(self) -> UserMessage:
        """|coro|

        Gets the original response for an interaction.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The fetched response!
        """
        if not self.has_replied:
            raise InteractionDoesNotExist(
                "No interaction reply has been sent yet!"
            )

        resp = await self._http.get(
            f"/webhooks/{self._client.bot.id}/{self.token}/messages/@original"
        )
        return UserMessage.from_dict(resp)

    async def ack(self, flags: Optional[InteractionFlags] = None):
        """|coro|

        Acknowledge an interaction, any flags here are applied to the reply.

        Parameters
        ----------
        flags :class:`~pincer.objects.app.interaction_flags.InteractionFlags`
            The flags which must be applied to the reply.

        Raises
        ------
        :class:`~pincer.exceptions.InteractionAlreadyAcknowledged`
            The interaction was already acknowledged, this can be
            because a reply or ack was already sent.
        """
        if self.has_replied or self.has_acknowledged:
            raise InteractionAlreadyAcknowledged(
                "The interaction you are trying to acknowledge has already "
                "been acknowledged"
            )

        self.has_acknowledged = True
        await self._http.post(
            f"interactions/{self.id}/{self.token}/callback",
            {
                "type": CallbackType.DEFERRED_MESSAGE,
                "data": {
                    "flags": flags
                }
            }
        )

    async def __post_send_handler(self, message: Message):
        """Process the interaction after it was sent.

        Parameters
        ----------
        message :class:`~pincer.objects.message.message.Message`
            The interaction message.
        """

        if message.delete_after:
            await sleep(message.delete_after)
            await self.delete()

    def __post_sent(self, message: Message):
        """Ensure the `__post_send_handler` method its future.

        Parameters
        ----------
        message :class:`~pincer.objects.message.message.Message`
            The interaction message.
        """
        self.has_replied = True
        ensure_future(self.__post_send_handler(message))

    async def reply(self, message: MessageConvertable):
        """|coro|

        Initial reply, only works if no ACK has been sent yet.

        Parameters
        ----------
        message :class:`~pincer.utils.convert_message.MessageConvertable`
            The response message!

        Raises
        ------
        :class:`~.pincer.errors.UseFollowup`
            Exception raised when a reply has already been sent so a
            :func:`~pincer.objects.app.interactions.Interaction.followup`
            should be used instead.
        :class:`~.pincer.errors.InteractionTimedOut`
            Exception raised when discord had to wait too long for a reply.
            You can extend the discord wait time by using the
            :func:`~pincer.objects.app.interaction.Interaction.ack`
            function.
        """
        if self.has_replied:
            raise UseFollowup(
                "A response has already been sent to the interaction. "
                "Please use a followup instead!"
            )
        elif self.has_acknowledged:
            self.has_replied = True
            await self.edit(message)
            return

        message = convert_message(self._client, message)
        content_type, data = message.serialize(
            message_type=CallbackType.MESSAGE
        )

        try:
            await self._http.post(
                f"interactions/{self.id}/{self.token}/callback",
                data,
                content_type=content_type
            )
        except NotFoundError:
            raise InteractionTimedOut(
                "Discord had to wait too long for the interaction reply, "
                "you can extend the time it takes for discord to timeout by "
                "acknowledging the interaction. (using interaction.ack)"
            )

        self.__post_sent(message)

    async def edit(self, message: MessageConvertable) -> UserMessage:
        """|coro|

        Edit an interaction. This is also the way to reply to
        interactions whom have been acknowledged.

        Parameters
        ----------
        message :class:`~pincer.utils.convert_message.MessageConvertable`
            The new message!

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The updated message object.

        Raises
        ------
        :class:`~.pincer.errors.InteractionDoesNotExist`
            Exception raised when no reply has been sent.
        """

        if not self.has_replied:
            raise InteractionDoesNotExist(
                "The interaction whom you are trying to edit has not "
                "been sent yet!"
            )

        message = convert_message(self._client, message)
        content_type, data = message.serialize()

        resp = await self._http.patch(
            f"webhooks/{self._client.bot.id}/{self.token}/messages/@original",
            data,
            content_type=content_type
        )
        self.__post_sent(message)
        return UserMessage.from_dict(resp)

    async def delete(self):
        """|coro|

        Delete the interaction.

        Raises
        ------
        :class:`~pincer.errors.InteractionDoesNotExist`
            Exception raised when no reply has been sent.
        """
        if not self.has_replied:
            raise InteractionDoesNotExist(
                "The interaction whom you are trying to delete has not "
                "been sent yet!"
            )

        await self._http.delete(
            f"webhooks/{self._client.bot.id}/{self.token}/messages/@original"
        )

    async def __post_followup_send_handler(
            self,
            followup: UserMessage,
            message: Message
    ):
        """Process a followup after it was sent.

        Parameters
        ----------
        followup :class:`~pincer.objects.message.user_message.UserMessage`
            The followup message that is being post processed.
        message :class:`~pincer.objects.message.message.Message`
            The followup message.
        """

        if message.delete_after:
            await sleep(message.delete_after)
            await self.delete_followup(followup.id)

    def __post_followup_sent(
            self,
            followup: UserMessage,
            message: Message
    ):
        """Ensure the `__post_followup_send_handler` method its future.

        Parameters
        ----------
        followup :class:`~pincer.objects.message.user_message.UserMessage`
            The followup message that is being post processed.
        message :class:`~pincer.objects.message.message.Message`
            The followup message.
        """
        ensure_future(self.__post_followup_send_handler(followup, message))

    async def followup(self, message: MessageConvertable) -> UserMessage:
        """|coro|

        Create a follow up message for the interaction.
        This allows you to respond with multiple messages.

        Parameters
        ----------
        message :class:`~pincer.utils.convert_message.MessageConvertable`
            The message to sent.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The message that has been sent.
        """
        message = convert_message(self._client, message)
        content_type, data = message.serialize()

        resp = await self._http.post(
            f"webhooks/{self._client.bot.id}/{self.token}",
            data,
            content_type=content_type
        )
        msg = UserMessage.from_dict(resp)
        self.__post_followup_sent(msg, message)
        return msg

    async def edit_followup(
            self,
            message_id: int,
            message: MessageConvertable
    ) -> UserMessage:
        """|coro|

        Edit a followup message.

        Parameters
        ----------
        message_id :class:`int`
            The id of the original followup message.
        message :class:`~pincer.utils.convert_message.MessageConvertable`
            The message new message.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The updated message object.
        """
        message = convert_message(self._client, message)
        content_type, data = message.serialize()

        resp = await self._http.patch(
            f"webhooks/{self._client.bot.id}/{self.token}/messages/{message_id}",
            data,
            content_type=content_type
        )
        msg = UserMessage.from_dict(resp)
        self.__post_followup_sent(msg, message)
        return msg

    async def get_followup(self, message_id: int) -> UserMessage:
        """|coro|

        Get a followup message by id.

        Parameters
        ----------
        message_id :class:`int`
            The id of the original followup message that must be fetched.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The fetched message object.
        """

        resp = await self._http.get(
            f"webhooks/{self._client.bot.id}/{self.token}/messages/{message_id}",
        )
        return UserMessage.from_dict(resp)

    async def delete_followup(self, message: Union[UserMessage, int]):
        """|coro|

        Remove a followup message by id.

        Parameters
        ----------
        message Union[:class:`~pincer.objects.user_message.UserMessage`, :class:`int`]
            The id/followup object of the followup message that must be deleted.
        """
        message_id = message if isinstance(message, int) else message.id

        await self._http.delete(
            f"webhooks/{self._client.bot.id}/{self.token}/messages/{message_id}",
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from typing import Any, Optional, Protocol, TypeVar

T = TypeVar("T")


class GetItem(Protocol):
    """Represents a class which implements the __getitem__ property."""
    def __getitem__(self, key: int) -> Any:
        return ...


def get_index(
        collection: GetItem,
        index: int,
        fallback: Optional[T] = None
) -> Optional[T]:
    """Gets an item from a collection through index.
    Allows you to provide a fallback for if that index is out of bounds.

    Parameters
    ----------
    collection: :class:`~pincer.utils.extraction.GetItem`
        The collection from which the item is retrieved.
    index: :class:`int`
        The index of the item in the collection.
    fallback: Optional[T]
        The fallback value which will be used if the index doesn't
        exist. Default value is None.

    Returns
    -------
    Optional[T]
        The item at the provided index from the collection, or if that
        item doesn't exist it will return the fallback value.
    """

    try:
        return collection[index]

    except IndexError:
        return fallback



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.


from __future__ import annotations

import logging
from asyncio import AbstractEventLoop, ensure_future, get_event_loop
from platform import system
from typing import Dict, Callable, Awaitable, Optional
from typing import TYPE_CHECKING

import zlib
from websockets import connect
from websockets.exceptions import ConnectionClosedError, ConnectionClosedOK
from websockets.legacy.client import WebSocketClientProtocol

from . import __package__
from .._config import GatewayConfig
from ..core.dispatch import GatewayDispatch
from ..core.heartbeat import Heartbeat
from ..exceptions import (
    PincerError, InvalidTokenError, UnhandledException,
    _InternalPerformReconnectError, DisallowedIntentsError
)

if TYPE_CHECKING:
    from ..objects.app.intents import Intents

ZLIB_SUFFIX = b'\x00\x00\xff\xff'

Handler = Callable[[WebSocketClientProtocol, GatewayDispatch], Awaitable[None]]
_log = logging.getLogger(__package__)


class Dispatcher:
    """The Dispatcher handles all interactions with discord websocket API.
    This also contains the main event loop, and handles the heartbeat.

    Running the dispatcher will create a connection with the
    Discord WebSocket API on behalf of the provided token.

    This token must be a bot token.
    (Which can be found on
    `<https://discord.com/developers/applications/>`_)
    """

    def __init__(
            self, token: str, *,
            handlers: Dict[int, Handler],
            intents: Intents,
            reconnect: bool
    ) -> None:
        if len(token) != 59:
            raise InvalidTokenError(
                "Discord Token must have exactly 59 characters."
            )

        self.__token = token
        self.__keep_alive = True
        self.__has_closed = self.__should_restart = False
        self.__socket: Optional[WebSocketClientProtocol] = None
        self.__intents = intents
        self.__reconnect = reconnect

        async def identify_and_handle_hello(
                socket: WebSocketClientProtocol,
                payload: GatewayDispatch
        ):
            """|coro|

            Identifies the client to the Discord Websocket API, this
            gets done when the client receives the ``hello`` (opcode 10)
            message from discord. Right after we send our identification
            the heartbeat starts.

            Parameters
            ----------
            socket : :class:`~ws:websockets.legacy.client.WebSocketClientProtocol`
                The current socket, which can be used to interact
                with the Discord API.
            payload : :class:`~pincer.core.dispatch.GatewayDispatch`
                The received payload from Discord.
            """  # noqa: E501
            _log.debug("Sending authentication/identification message.")

            await socket.send(self.__hello_socket)
            await Heartbeat.handle_hello(socket, payload)

        async def handle_reconnect(_, payload: GatewayDispatch):
            """|coro|

            Closes the client and then reconnects it.
            """
            _log.debug("Reconnecting client...")
            await self.restart(payload.seq)

        self.__dispatch_handlers: Dict[int, Handler] = {
            **handlers,
            7: handle_reconnect,
            9: handle_reconnect,
            10: identify_and_handle_hello,
            11: Heartbeat.handle_heartbeat
        }

        self.__dispatch_errors: Dict[int, PincerError] = {
            1006: _InternalPerformReconnectError(),
            4000: _InternalPerformReconnectError(),
            4004: InvalidTokenError(),
            4007: _InternalPerformReconnectError(),
            4009: _InternalPerformReconnectError(),
            4014: DisallowedIntentsError()
        }

    @property
    def intents(self):
        """:class:`app.Intents`"""
        return self.__intents

    @property
    def __hello_socket(self) -> str:
        return str(
            GatewayDispatch(
                2, {
                    "token": self.__token,
                    "intents": self.__intents,
                    "properties": {
                        "$os": system(),
                        "$browser": __package__,
                        "$device": __package__
                    },
                    "compress": GatewayConfig.compressed()
                }
            )
        )

    async def __handler_manager(
            self,
            socket: WebSocketClientProtocol,
            payload: GatewayDispatch,
            loop: AbstractEventLoop
    ):
        """
        This manages all handles for given OP codes.
        This method gets invoked for every message that is received from
        Discord.

        :param socket:
            The current socket, which can be used to interact with
            the Discord API.
        :param payload:
            The received payload from Discord.
        :param loop:
            The current async loop on which the future is bound.
        """
        _log.debug(
            "New event received, checking if handler exists for opcode: %i",
            payload.op
        )

        handler: Handler = self.__dispatch_handlers.get(payload.op)
        all_handler: Handler = self.__dispatch_handlers.get(-1)

        if not handler:
            _log.error(
                "No handler was found for opcode %i, please report this to the "
                "pincer dev team!", payload.op
            )

            if not all_handler:
                raise UnhandledException(f"Unhandled payload: {payload}")

        _log.debug(
            "Event handler found, ensuring async future in current loop."
        )

        def execute_handler(event_handler: Handler):
            if event_handler:
                ensure_future(event_handler(socket, payload), loop=loop)

        execute_handler(handler)
        execute_handler(all_handler)

    async def __dispatcher(self, loop: AbstractEventLoop):
        """
        The main event loop.
        This handles all interactions with the websocket API.

        :param loop:
            The loop in which the dispatcher is running.
        """
        _log.debug(
            "Establishing websocket connection with `%s`", GatewayConfig.uri()
        )

        async with connect(GatewayConfig.uri()) as socket:
            socket: WebSocketClientProtocol = socket
            self.__socket = socket

            # Removing the limit of the received socket.
            # Having the default limit can cause an issue
            # with first payload of bigger bots.
            socket.max_size = None

            _log.debug(
                "Successfully established websocket connection with `%s`",
                GatewayConfig.uri()
            )

            if GatewayConfig.compression == "zlib-stream":
                # Create an inflator for compressed data as defined in
                # https://discord.com/developers/docs/topics/gateway
                inflator = zlib.decompressobj()

            while self.__keep_alive:
                try:
                    _log.debug("Waiting for new event.")
                    msg = await socket.recv()

                    if msg == "CLOSE":
                        break

                    if isinstance(msg, bytes):
                        if GatewayConfig.compression == "zlib-payload":
                            msg = zlib.decompress(msg)
                        else:
                            buffer = bytearray(msg)

                            while not buffer.endswith(ZLIB_SUFFIX):
                                buffer.extend(await socket.recv())

                            msg = inflator.decompress(buffer).decode('utf-8')

                    await self.__handler_manager(
                        socket,
                        GatewayDispatch.from_string(msg),
                        loop
                    )

                except ConnectionClosedError as exc:
                    _log.debug(
                        "The connection with `%s` has been broken unexpectedly."
                        " (%i, %s)", GatewayConfig.uri(), exc.code, exc.reason
                    )

                    await self.close()
                    exception = self.__dispatch_errors.get(exc.code)

                    if (
                        isinstance(exception, _InternalPerformReconnectError)
                        and self.__reconnect
                    ):
                        _log.debug("Connection closed, reconnecting...")
                        return await self.restart()

                    raise exception or UnhandledException(
                        f"Dispatch error ({exc.code}): {exc.reason}"
                    )
                except ConnectionClosedOK:
                    _log.debug("Connection closed successfully.")
            self.__has_closed = True

    def start_loop(self, *, loop: AbstractEventLoop = None):
        """Instantiate the dispatcher, this will create a connection to the
        Discord websocket API on behalf of the client who's token has
        been passed.

        Parameters
        ----------
        loop : :class:`~asyncio.AbstractEventLoop`
            The loop in which the Dispatcher will run. If no loop is
            provided it will get a new one. |default| :data:`None`
        """
        loop = loop or get_event_loop()
        self.__keep_alive = True
        self.__has_closed = self.__should_restart = False
        loop.run_until_complete(self.__dispatcher(loop))
        if self.__should_restart:
            return self.start_loop(loop=loop)
        loop.close()

    async def restart(self, /, seq: Optional[int] = None):
        """
        Restart the dispatcher.

        Parameters
        ----------
        seq Optional[:class:`int`]
            The sequence number of the last dispatched event.
            If not provided, the dispatcher will restart with no base
            sequence.
        """
        await self.close()
        Heartbeat.update_sequence(seq)
        self.__should_restart = True

    async def close(self):
        """|coro|

        Stop the dispatcher from listening and responding to gateway
        events. This should let the client close on itself.
        """
        if not self.__socket:
            _log.error("Cannot close non existing socket socket connection.")
            raise RuntimeError("Please open the connection before closing.")

        _log.debug("Closing connection...")
        self.__keep_alive = False

        self.__socket.messages.append("CLOSE")
        if waiter := getattr(self.__socket, "_pop_message_waiter", None):
            waiter.cancel()
        await self.__socket.close()
        _log.debug("Successfully closed connection!")



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Optional, TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from .user import User
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp


class IntegrationExpireBehavior(IntEnum):
    """Represents a Discord Integration expire behavior

    Attributes
    ----------
    REMOVE_ROLE:
        Remove role on expire.
    KICK:
        Kick on expire.
    """
    REMOVE_ROLE = 0
    KICK = 1


@dataclass
class IntegrationAccount(APIObject):
    """Represents a Discord Integration Account object

    Attributes
    ----------
    id: :class:`str`
        Id of the account
    name: :class:`str`
        Name of the account
    """
    id: str
    name: str


@dataclass
class IntegrationApplication(APIObject):
    """Represents a Discord Integration Application object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of the app
    name: :class:`str`
        The name of the app
    icon: Optional[:class:`str`]
        The icon hash of the app
    description: :class:`str`
        The description of the app
    summary: :class:`str`
        The summary of the app
    bot: APINullable[:class:`~pincer.objects.user.user.User`]
        The bot associated with this application
    """
    id: Snowflake
    name: str
    icon: Optional[str]
    description: str
    summary: str
    bot: APINullable[User] = MISSING


@dataclass
class Integration(APIObject):
    """Represents a Discord Integration object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Integration id
    name: :class:`str`
        Integration name
    type: :class:`str`
        Integration type (twitch, youtube, or discord)$
    enabled: :class:`bool`
        Is this integration enabled
    account: :class:`~pincer.objects.user.integration.IntegrationAccount`
        Integration account information
    syncing: APINullable[:class:`bool`]
        Is this integration syncing
    role_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id that this integration uses for subscribers
    enable_emoticons: APINullable[:class:`bool`]
        Whether emoticons should be synced for this integration
        (twitch only currently)
    expire_behavior: APINullable[:class:`~pincer.objects.user.integration.IntegrationExpireBehavior`]
        The behavior of expiring subscribers
    expire_grace_period: APINullable[:class:`int`]
        The grace period (in days) before expiring subscribers
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        User for this integration
    synced_at: APINullable[:class:`~pincer.utils.timestamp.Timestamp`]
        When this integration was last synced
    subscriber_count: APINullable[:class:`int`]
        How many subscribers this integration has
    revoked: APINullable[:class:`bool`]
        Has this integration been revoked
    application: APINullable[:class:`~pincer.objects.user.integration.IntegrationApplication`]
        The bot/OAuth2 application for discord integrations
    """
    # noqa: E501

    id: Snowflake
    name: str
    type: str
    enabled: bool
    account: IntegrationAccount

    syncing: APINullable[bool] = MISSING
    role_id: APINullable[Snowflake] = MISSING
    enable_emoticons: APINullable[bool] = MISSING
    expire_behavior: APINullable[IntegrationExpireBehavior] = MISSING
    expire_grace_period: APINullable[int] = MISSING
    user: APINullable[User] = MISSING
    synced_at: APINullable[Timestamp] = MISSING
    subscriber_count: APINullable[int] = MISSING
    revoked: APINullable[bool] = MISSING
    application: APINullable[IntegrationApplication] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a thread is updated"""

from ..core.dispatch import GatewayDispatch
from ..objects import Channel
from ..utils.conversion import construct_client_dict


async def thread_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_thread_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the thread update event.


    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.channel.Channel`]]
        ``on_thread_update`` and an ``Channel``
    """

    return "on_thread_update", [
        Channel.from_dict(construct_client_dict(self, payload.data))
    ]


def export():
    return thread_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a channel is updated"""

from ..core.dispatch import GatewayDispatch
from ..objects import Channel
from ..utils.conversion import construct_client_dict


async def channel_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_channel_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the channel update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.channel.channel.Channel`]]
        ``on_channel_update`` and a ``Channel``
    """

    channel = Channel.from_dict(construct_client_dict(self, payload.data))

    if channel.guild_id in self.guilds.keys():
        guild = self.guilds[channel.guild_id]
        old = filter(lambda c: c.id == channel.id, guild.channels)
        if old:
            guild.channels.remove(old)
        guild.channels.append(channel)

    return "on_channel_update", [channel]


def export():
    return channel_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""Event sent when a subscribed server's state changes"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.guild import GuildStatusEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def guild_status_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_status`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild status event.

    Return
    ------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.guild.GuildStatusEvent`]]
        ``on_guild_status`` and a ``GuildStatusEvent``
    """
    return "on_guild_status", [
        GuildStatusEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return guild_status_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from datetime import datetime
from typing import Optional, TypeVar, Union

DISCORD_EPOCH = 1420070400
TS = TypeVar("TS", str, datetime, float, int)


class Timestamp:
    """Contains a lot of useful methods for working with timestamps.

    Attributes
    ----------
    date: :class:`str`
        The time of the timestamp.
    time: :class:`str`
        Alias for date.

    Parameters
    ----------
    time: Union[:class:`str`, :class:`int`, :class:`float`, :class:`datetime.datetime`]
    """  # noqa: E501

    def __init__(self, time: Optional[TS] = None):
        self.__time = Timestamp.parse(time)
        self.__epoch = Timestamp.to_epoch(self.__time)
        self.date, self.time = str(self).split()

    @classmethod
    def __factory__(cls, time: Optional[TS] = None) -> datetime:
        return cls.parse(time)

    @staticmethod
    def to_epoch(time: datetime) -> int:
        """Convert a datetime to an epoch.

        time: :class:`datetime.datetime`
            The datetime to convert.

        Returns
        -------
        :class:`int`:
            The epoch time integer.
        """
        return int(time.timestamp() * 1000)

    @staticmethod
    def string_to_datetime(string: str) -> datetime:
        """Convert a string to a datetime object.

        string: :class:`str`
            The string to convert.

        Returns
        -------
        :class:`datetime.datetime`:
            The converted datetime object.
        """
        return datetime.fromisoformat(string)

    @staticmethod
    def epoch_to_datetime(epoch: Union[int, float]) -> datetime:
        """Convert an epoch to a datetime object.

        epoch: Union[:class:`int`, :class:`float`]
            The epoch to convert to a datetime object.

        Returns
        -------
        :class:`datetime.datetime`:
            The converted datetime object.
        """
        return datetime.utcfromtimestamp(epoch)

    @staticmethod
    def parse(time: Optional[TS] = None) -> datetime:
        """Convert a time to datetime object.

        time: Optional[Union[:class:`str`, :class:`int`, :class:`float`, :class:`datetime.datetime`]]
            The time to be converted to a datetime object.
            This can be one of these types: datetime, float, int, str

            If no parameter is passed it will return the current
            datetime.

        Returns
        -------
        :class:`datetime.datetime`:
            The converted datetime object.
        """  # noqa: E501

        if isinstance(time, datetime):
            return time

        elif isinstance(time, str):
            return Timestamp.string_to_datetime(time)

        elif isinstance(time, (int, float)):
            dt = Timestamp.epoch_to_datetime(t := int(time))

            if dt.year < 2015:  # if not Discord Epoch
                t += DISCORD_EPOCH  # make it Unix Epoch
            return Timestamp.epoch_to_datetime(t)

        return datetime.now()

    def __getattr__(self, key: str) -> int:
        return getattr(self.__time, key)

    def __str__(self) -> str:
        if len(string := str(self.__time)) == 19:
            return string + ".000"
        return string[:-3]

    def __int__(self) -> int:
        return self.__epoch  # in milliseconds

    def __float__(self) -> float:
        return self.__epoch / 1000  # in seconds

    def __ge__(self, other: Timestamp) -> bool:
        return self.__epoch >= other.__epoch

    def __gt__(self, other: Timestamp) -> bool:
        return self.__epoch > other.__epoch

    def __le__(self, other: Timestamp) -> bool:
        return self.__epoch <= other.__epoch

    def __lt__(self, other: Timestamp) -> bool:
        return self.__epoch < other.__epoch

    def __eq__(self, other: Timestamp) -> bool:
        return self.__epoch == other.__epoch

    def __ne__(self, other: Timestamp) -> bool:
        return self.__epoch != other.__epoch



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Optional, List

    from ..user.user import User
    from ..guild.role import Role
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


@dataclass
class Emoji(APIObject):
    """Representation of an emoji in a class.

    Attributes
    ----------
    id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        Emoji id
    name: Optional[:class:`str`]
        Emoji name
    animated: APINullable[:class:`bool`]
        Whether this emoji is animated
    available: APINullable[:class:`bool`]
        Whether this emoji can be used, may be false due to loss of Server
        Boosts
    managed: APINullable[:class:`bool`]
        Whether this emoji is managed
    require_colons: APINullable[:class:`bool`]
        Whether this emoji must be wrapped in colons
    roles: APINullable[List[:class:`~pincer.objects.guild.role.Role`]]
        Roles allowed to use this emoji
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        User that created this emoji
    """

    name: Optional[str]
    id: APINullable[Snowflake] = MISSING

    animated: APINullable[bool] = MISSING
    available: APINullable[bool] = MISSING
    managed: APINullable[bool] = MISSING
    require_colons: APINullable[bool] = MISSING
    roles: APINullable[List[Role]] = MISSING
    user: APINullable[User] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild is updated"""
from typing import List

from ..core.dispatch import GatewayDispatch
from ..objects import Guild, Channel
from ..utils.conversion import construct_client_dict


async def guild_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.guild.Guild`]]
        ``on_guild_Update`` and an ``Guild``
    """

    channel_list = payload.data.pop("channels", [])

    channels: List[Channel] = [
        Channel.from_dict(construct_client_dict(self, channel))
        for channel in channel_list
    ]

    guild = Guild.from_dict(construct_client_dict(
        self,
        {"channels": channels, **payload.data}
    ))
    self.guild[guild.id] = guild

    return "on_guild_update", [guild]


def export():
    return guild_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import auto, Enum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import APINullable, MISSING

if TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Tuple, Union

    from .presence import Activity
    from ..app.intents import Intents
    from ...utils.snowflake import Snowflake


@dataclass
class Identify(APIObject):
    """Used to trigger the initial handshake with the gateway.

    Attributes
    ----------
    token: :class:`str`
        Authentication token
    properties: Dict[:class:`str`, :class:`str`]
        Connection properties
    intents: :class:`~pincer.objects.app.intents.Intents`
        The Gateway Intents you wish to receive
    compress: APINullable[:class:`bool`]
        Whether this connection supports compression of packets
    large_threshold: APINullable[:class:`int`]
        Value between 50 and 250, total number
        of members where the gateway will stop sending offline
        members in the guild member list
    shard: APINullable[Tuple[:class:`int`, :class:`int`]]
        Used for Guild Sharding
    presence: APINullable[Any]
        Presence structure for initial presence information
    """
    token: str
    properties: Dict[str, str]
    intents: Intents

    compress: APINullable[bool] = MISSING
    large_threshold: APINullable[int] = MISSING
    shard: APINullable[Tuple[int, int]] = MISSING
    presence: APINullable[Any] = MISSING  # FIXME


@dataclass
class Resume(APIObject):
    """Used to replay missed events when a disconnected client resumes.

    Attributes
    ----------
    token: :class:`str`
        Session token
    session_id: :class:`str`
        Session id
    seq: :class:`int`
        Last sequence number received
    """
    token: str
    session_id: str
    seq: int


@dataclass
class RequestGuildMembers(APIObject):
    """Used to request all members for a guild or a list of guilds.

    guild_id:
        id of the guild to get members for

    query:
        string that username starts with, or an empty string
        to return all members

    limit:
        maximum number of members to send matching the `query`;
        a limit of `0` can be used with an empty string `query`
        to return all members

    presences:
        used to specify if we want the presences of the matches members

    user_ids:
        used to specify which users you wish to fetch

    nonce:
        nonce to identify the Guild Members Chunk response
    """
    guild_id: Snowflake
    limit: int

    query: APINullable[str] = MISSING
    presences: APINullable[bool] = MISSING
    user_ids: APINullable[Union[Snowflake, List[Snowflake]]] = MISSING
    nonce: APINullable[str] = MISSING


@dataclass
class UpdateVoiceState(APIObject):
    """Sent when a client wants to join, move,
    or disconnect from a voice channel.

    guild_id:
        id of the guild

    channel_id:
        id of the voice channel client
        wants to join (null if disconnecting)

    self_mute:
        is the client muted

    self_deaf:
        is the client deafened
    """
    guild_id: Snowflake
    self_mute: bool
    self_deaf: bool

    channel_id: Optional[Snowflake] = None


class StatusType(Enum):
    """online:
        Online

    dnd:
        Do Not Disturb

    idle:
        AFK

    invisible:
        Invisible and shown as offline

    offline:
        Offline
    """
    online = auto()
    dnd = auto()
    idle = auto()
    invisible = auto()
    offline = auto()


@dataclass
class UpdatePresence(APIObject):
    """Sent by the client to indicate a presence or status update.

    since:
        unix time (in milliseconds) of when the client went idle,
        or null if the client is not idle

    activities:
        the user's activities

    status:
        the user's new status

    afk:
        whether or not the client is afk
    """
    activities: List[Activity]
    status: StatusType
    afk: bool
    since: Optional[int] = None



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, IntEnum
from typing import TYPE_CHECKING

from .attachment import Attachment
from .component import MessageComponent
from .embed import Embed
from .reaction import Reaction
from .reference import MessageReference
from .sticker import StickerItem
from ..app.application import Application
from ..app.interaction_base import MessageInteraction
from ..guild.member import GuildMember
from ..guild.role import Role
from ..user.user import User
from ..._config import GatewayConfig
from ...utils.api_object import APIObject
from ...utils.conversion import construct_client_dict
from ...utils.snowflake import Snowflake
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Any, List, Optional, Union, Generator

    from ..guild.channel import Channel, ChannelMention
    from ...utils.types import APINullable
    from ...utils.timestamp import Timestamp


class AllowedMentionTypes(str, Enum):
    """The allowed mentions.

    Attributes
    ----------
    ROLES:
        Controls role mentions
    USERS:
        Controls user mentions
    EVERYONE:
        Controls @everyone and @here mentions
    """
    ROLES = "roles"
    USERS = "users"
    EVERYONE = "everyone"


@dataclass
class AllowedMentions(APIObject):
    """Represents the entities the client can mention

    Attributes
    ----------
    parse: List[:class:`~pincer.objects.message.user_message.AllowedMentionTypes`]
        An array of allowed mention types to parse from the content.
    roles: List[Union[:class:`~pincer.objects.guild.role.Role`, :class:`~pincer.utils.snowflake.Snowflake`]]
        List of ``Role`` objects or snowflakes of allowed mentions.
    users: List[Union[:class:`~pincer.objects.user.user.User` :class:`~pincer.utils.snowflake.Snowflake`]]
        List of ``user`` objects or snowflakes of allowed mentions.
    reply: :class:`bool`
        If replies should mention the author.
        |default| :data:`True`
    """  # noqa: E501

    parse: List[AllowedMentionTypes]
    roles: List[Union[Role, Snowflake]]
    users: List[Union[User, Snowflake]]
    reply: bool = True

    def to_dict(self):
        def get_str_id(obj: Union[Snowflake, User, Role]) -> str:
            if hasattr(obj, "id"):
                obj = obj.id

            return str(obj)

        return {
            "parse": self.parse,
            "roles": list(map(get_str_id, self.roles)),
            "users": list(map(get_str_id, self.users)),
            "replied_user": self.reply
        }


class MessageActivityType(IntEnum):
    """The activity people can perform on a rich presence activity.

    Such an activity could for example be a spotify listen.

    Attributes
    ----------
    JOIN:
        Invite to join.
    SPECTATE:
        Invite to spectate.
    LISTEN:
        Invite to listen along.
    JOIN_REQUEST:
        Request to join.
    """
    JOIN = 1
    SPECTATE = 2
    LISTEN = 3
    JOIN_REQUEST = 5


class MessageFlags(IntEnum):
    """Special message properties.

    Attributes
    ----------
    CROSSPOSTED:
        The message has been published to subscribed
        channels (via Channel Following)
    IS_CROSSPOST:
        This message originated from a message
        in another channel (via Channel Following)
    SUPPRESS_EMBEDS:
        Do not include any embeds when serializing this message
    SOURCE_MESSAGE_DELETED:
        The source message for this crosspost
        has been deleted (via Channel Following)
    URGENT:
        This message came from the urgent message system
    HAS_THREAD:
        This message has an associated thread,
        with the same id as the message
    EPHEMERAL:
        This message is only visible to the user
        who invoked the Interaction
    LOADING:
        This message is an Interaction
        Response and the bot is "thinking"
    """
    CROSSPOSTED = 1 << 0
    IS_CROSSPOST = 1 << 1
    SUPPRESS_EMBEDS = 1 << 2
    SOURCE_MESSAGE_DELETED = 1 << 3
    URGENT = 1 << 4
    HAS_THREAD = 1 << 5
    EPHEMERAL = 1 << 6
    LOADING = 1 << 7


class MessageType(IntEnum):
    """Represents the type of the message.

    Attributes
    ----------
    DEFAULT:
        Normal message.
    RECIPIENT_ADD:
        Recipient is added.
    RECIPIENT_REMOVE:
        Recipient is removed.
    CALL:
        A call is being made.
    CHANNEL_NAME_CHANGE:
        The group channel name is changed.
    CHANNEL_ICON_CHANGE:
        The group channel icon is changed.
    CHANNEL_PINNED_MESSAGE:
        A message is pinned.
    GUILD_MEMBER_JOIN:
        A member joined.
    USER_PREMIUM_GUILD_SUBSCRIPTION:
        A boost.
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_1:
        A boost that reached tier 1.
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_2:
        A boost that reached tier 2.
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_3:
        A boost that reached tier 3.
    CHANNEL_FOLLOW_ADD:
        A channel is subscribed to.
    GUILD_DISCOVERY_DISQUALIFIED:
        The guild is disqualified from discovery,
    GUILD_DISCOVERY_REQUALIFIED:
        The guild is requalified for discovery.
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING:
        Warning about discovery violations.
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING:
        Final warning about discovery violations.
    THREAD_CREATED:
        A thread is created.
    REPLY:
        A message reply.
    APPLICATION_COMMAND:
        Slash command is used and responded to.
    THREAD_STARTER_MESSAGE:
        The initial message in a thread when its created off a message.
    GUILD_INVITE_REMINDER:
        ??
    """
    DEFAULT = 0
    RECIPIENT_ADD = 1
    RECIPIENT_REMOVE = 2
    CALL = 3
    CHANNEL_NAME_CHANGE = 4
    CHANNEL_ICON_CHANGE = 5
    CHANNEL_PINNED_MESSAGE = 6
    GUILD_MEMBER_JOIN = 7
    USER_PREMIUM_GUILD_SUBSCRIPTION = 8
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_1 = 9
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_2 = 10
    USER_PREMIUM_GUILD_SUBSCRIPTION_TIER_3 = 11
    CHANNEL_FOLLOW_ADD = 12
    GUILD_DISCOVERY_DISQUALIFIED = 14
    GUILD_DISCOVERY_REQUALIFIED = 15
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    THREAD_CREATED = 18
    REPLY = 19
    APPLICATION_COMMAND = 20

    if GatewayConfig.version < 8:
        REPLY = 0
        APPLICATION_COMMAND = 0

    if GatewayConfig.version >= 9:
        THREAD_STARTER_MESSAGE = 21

    GUILD_INVITE_REMINDER = 22


@dataclass
class MessageActivity(APIObject):
    """Represents a Discord Message Activity object

    Attributes
    ----------
    type: :class:`~pincer.objects.message.user_message.MessageActivity`
        type of message activity
    party_id: APINullable[:class:`str`]
        party_id from a Rich Presence event
    """
    type: MessageActivityType
    party_id: APINullable[str] = MISSING


@dataclass
class UserMessage(APIObject):
    """Represents a message sent in a channel within Discord.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Ud of the message
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the channel the message was sent in
    author: :class:`~pincer.objects.user.user.User`
        The author of this message (not guaranteed to be a valid user)
    content: :class:`str`
        Contents of the message
    timestamp: :class:`~pincer.utils.timestamp.Timestamp`
        When this message was sent
    edited_timestamp: Optional[:class:`~pincer.utils.timestamp.Timestamp`]
        When this message was edited (or null if never)
    tts: :class:`bool`
        Whether this was a TTS message
    mention_everyone: :class:`bool`
        Whether this message mentions everyone
    mentions: List[:class:`~pincer.objects.guild.member.GuildMember`]
        Users specifically mentioned in the message
    mention_roles: List[:class:`~pincer.objects.guild.role.Role`]
        Roles specifically mentioned in this message
    attachments: List[:class:`~pincer.objects.message.attachment.Attachment`]
        Any attached files
    embeds: List[:class:`~pincer.objects.message.embed.Embed`]
        Any embedded content
    pinned: :class:`bool`
        Whether this message is pinned
    type: :class:`~pincer.objects.message.user_message.MessageType`
        Type of message
    mention_channels: APINullable[List[:class:`~pincer.objects.guild.channel.Channel`]]
        Channels specifically mentioned in this message
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the guild the message was sent in
    member: APINullable[:class:`~pincer.objects.guild.member.PartialGuildMember`]
        Member properties for this message's author
    reactions: APINullable[List[:class:`~pincer.objects.message.reaction.Reaction`]]
        Reactions to the message
    nonce: APINullable[Union[:class:`int`, :class:`str`]]
        User for validating a message was sent
    webhook_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        If the message is generated by a webhook,
        this is the webhook's id
    activity: APINullable[:class:`~pincer.objects.message.user_message.MessageActivity`]
        Sent with Rich Presence-related chat embeds
    application: APINullable[:class:`~pincer.objects.app.application.Application`]
        Sent with Rich Presence-related chat embeds
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        If the message is a response to an Interaction,
        this is the id of the interaction's application
    message_reference: APINullable[:class:`~pincer.objects.message.reference.MessageReference`]
        Data showing the source of a crosspost,
        channel follow add, pin, or reply message
    flags: APINullable[:class:`~pincer.objects.message.user_message.MessageFlags`]
        Message flags combined as a bitfield
    referenced_message: APINullable[Optional[:class:`~pincer.objects.message.user_message.UserMessage`]]
        The message associated with the message_reference
    interaction: APINullable[:class:`~pincer.objects.app.interaction_base.MessageInteraction`]
        Sent if the message is a response to an Interaction
    thread: APINullable[:class:`~pincer.objects.guild.channel.Channel`]
        The thread that was started from this message,
        includes thread member object
    components: APINullable[List[:class:`~pincer.objects.message.component.MessageComponent`]]
        Sent if the message contains components like buttons,
        action rows, or other interactive components
    sticker_items: APINullable[List[:class:`~pincer.objects.message.sticker.StickerItem`]]
        Sent if the message contains stickers
    """
    # noqa: E501

    id: Snowflake
    channel_id: Snowflake
    author: User
    content: str
    timestamp: Timestamp
    tts: bool
    mention_everyone: bool
    mentions: List[GuildMember]
    mention_roles: List[Role]
    attachments: List[Attachment]
    embeds: List[Embed]
    pinned: bool
    type: MessageType

    edited_timestamp: APINullable[Timestamp] = MISSING
    mention_channels: APINullable[List[ChannelMention]] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING
    reactions: APINullable[List[Reaction]] = MISSING
    nonce: APINullable[Union[int, str]] = MISSING
    webhook_id: APINullable[Snowflake] = MISSING
    activity: APINullable[MessageActivity] = MISSING
    application: APINullable[Application] = MISSING
    application_id: APINullable[Snowflake] = MISSING
    message_reference: APINullable[MessageReference] = MISSING
    flags: APINullable[MessageFlags] = MISSING
    referenced_message: APINullable[Optional[UserMessage]] = MISSING
    interaction: APINullable[MessageInteraction] = MISSING
    thread: APINullable[Channel] = MISSING
    components: APINullable[List[MessageComponent]] = MISSING
    sticker_items: APINullable[List[StickerItem]] = MISSING

    def __str__(self):
        return self.content

    async def get_most_recent(self):
        """|coro|

        Certain Discord methods don't return the message object data after its
        updated. This function can be run to get the most recent version of the
        message object.
        """

        return self.from_dict(
            construct_client_dict(
                self._client,
                await self._http.get(
                    f"/channels/{self.channel_id}/messages/{self.id}"
                )
            )
        )

    async def react(self, emoji: str):
        """|coro|

        Create a reaction for the message. Requires the
        ``READ_MESSAGE_HISTORY` intent. ``ADD_REACTIONS`` intent is required if
        nobody else has reacted using the emoji.

        Parameters
        ----------
        emoji: :class:`str`
            Character for emoji. Does not need to be URL encoded.
        """

        await self._http.put(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}/@me"
        )

    async def unreact(self, emoji: str):
        """|coro|

        Delete a reaction the current user has made for the message.

        Parameters
        ----------
        emoji: :class:`str`
            Character for emoji. Does not need to be URL encoded.
        """

        await self._http.delete(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}/@me"
        )

    async def remove_user_reaction(self, emoji: str, user_id: Snowflake):
        """|coro|

        Deletes another user's reaction. Requires the ``MANAGE_MESSAGES``
        intent.

        Parameters
        ----------
        emoji: :class:`str`
            Character for emoji. Does not need to be URL encoded.
        user_id: :class:`~pincer.utils.snowflake.Snowflake`
            User ID
        """

        await self._http.delete(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}"
            f"/{user_id}"
        )

    async def get_reactions(
        self, emoji: str, after: Snowflake = 0, limit=25
    ) -> Generator[User, None, None]:
        # TODO: HTTP Client will need to refactored to allow parameters using aiohttp's system.
        """|coro|

        Returns the users that reacted with this emoji.

        Parameters
        ----------
        emoji: :class:`str`
            Emoji to get users for.
        after: :class:`~pincer.utils.snowflake.Snowflake`
            Get users after this user ID. Returns all users if not provided.
            |default| ``0``
        limit: :class:`int`
            Max number of users to return (1-100).
            |default| ``25``
        """

        for user in await self._http.get(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}"
            f"?after={after}&limit={limit}"
        ):
            yield User.from_dict(user)

    async def remove_all_reactions(self):
        """|coro|

        Delete all reactions on a message. Requires the ``MANAGE_MESSAGES``
        intent.
        """

        await self._http.delete(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions"
        )

    async def remove_emoji(self, emoji):
        """|coro|

        Deletes all the reactions for a given emoji on a message. Requires the
        ``MANAGE_MESSAGES`` intent.

        Parameters
        ----------
        emoji: :class:`str`
            Character for emoji. Does not need to be URL encoded.
        """

        await self._http.delete(
            f"/channels/{self.channel_id}/messages/{self.id}/reactions/{emoji}"
        )

    # TODO: Implement file (https://discord.com/developers/docs/resources/channel#edit-message)
    async def edit(
        self,
        content: str = None,
        embeds: List[Embed] = None,
        flags: int = None,
        allowed_mentions: AllowedMentions = None,
        attachments: List[Attachment] = None,
        components: List[MessageComponent] = None
    ):
        """|coro|

        Edit a previously sent message. The fields content, embeds, and flags
        can be edited by the original message author. Other users can only
        edit flags and only if they have the ``MANAGE_MESSAGES`` permission in
        the corresponding channel. When specifying flags, ensure to include
        all previously set flags/bits in addition to ones that you are
        modifying.

        Parameters
        ----------
        content: :class:`str`
            The message contents (up to 2000 characters)
            |default| ``None``
        embeds: List[:class:`~pincer.objects.message.embed.Embed`]
            Embedded rich content (up to 6000 characters)
        flags: :class:`int`
            Edit the flags of a message (only ``SUPPRESS_EMBEDS`` can
            currently be set/unset)
        allowed_mentions: :class:`~pincer.objects.message.message.AllowedMentions`
            allowed mentions for the message
        attachments: List[:class:`~pincer.objects.message.attachment.Attachment`]
            attached files to keep
        components: List[:class:`~pincer.objects.message.component.MessageComponent`]
            the components to include with the message
        """

        data = {}

        def set_if_not_none(value: Any, name: str):
            if isinstance(value, APIObject):
                data[name] = value.to_dict()
            elif value is not None:
                data[name] = value

        set_if_not_none(content, "content")
        set_if_not_none(embeds, "embeds")
        set_if_not_none(flags, "flags")
        set_if_not_none(allowed_mentions, "allowed_mentions")
        set_if_not_none(attachments, "attachments")
        set_if_not_none(components, "components")

        await self._http.patch(
            f"/channels/{self.channel_id}/messages/{self.id}",
            data=data
        )

    async def delete(self):
        """|coro|

        Delete a message. Requires the ``MANAGE_MESSAGES`` intent if the
        message was not sent by the current user.
        """

        await self._http.delete(
            f"/channels/{self.channel_id}/messages/{self.id}"
        )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ..user.user import User
from ...utils.api_object import APIObject
from ...utils.snowflake import Snowflake
from ...utils.timestamp import Timestamp
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List, Optional

    from ...client import Client
    from ...utils.types import APINullable


@dataclass
class BaseMember(APIObject):
    """Represents the base of a guild member.

    Attributes
    ----------
    deaf: :class:`bool`
        Whether the user is deafened in voice channels
    joined_at: :class:`~pincer.utils.timestamp.Timestamp`
        Then the user joined the guild
    mute: :class:`bool`
        Whether the user is muted in voice channels
    roles: List[:class:`~pincer.utils.snowflake.Snowflake`]
        Array of role object ids
    hoisted_role: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The user's top role in the guild.
    """
    deaf: bool
    joined_at: Timestamp
    mute: bool
    roles: List[Snowflake]

    hoisted_role: APINullable[Snowflake] = MISSING


@dataclass
class PartialGuildMember(APIObject):
    """Represents a partial guild member.
    This is a reference to a member from a guild which does not contain
    all information.

    This gets used in form example message mentions.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The user's id
    username: :class:`str`
        The user's username, not unique across the platform
    discriminator: :class:`str`
        The user's 4-digit discord-tag
    avatar: :class:`str`
        The user's avatar hash
    public_flags: :class:`int`
        The flags on a user's account
    member: :class:`~pincer.objects.guild.member.BaseMember`
        The user their (partial) guild information.
    """
    id: Snowflake
    username: str
    discriminator: str
    avatar: str
    public_flags: int
    member: Optional[BaseMember]


@dataclass
class GuildMember(BaseMember, APIObject):
    """Represents a member which resides in a guild/server.

    Attributes
    ----------
    nick: APINullable[Optional[:class:`str`]]
        This users guild nickname
    pending: APINullable[:class:`bool`]
        Whether the user has not yet passed the guild's Membership
        Screening requirements
    is_pending: APINullable[:class:`bool`]
        Deprecated version of pending.
    permissions: APINullable[:class:`str`]
        Total permissions of the member in the channel,
        including overwrites, returned when in the interaction object
    premium_since: APINullable[Optional[:class:`~pincer.utils.timestamp.Timestamp`]]
        When the user started boosting the guild
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        The user this guild member represents
    Avatar: APINullable[:class:`str`]
        The user's avatar.
    """  # noqa: E501

    nick: APINullable[Optional[str]] = MISSING
    pending: APINullable[bool] = MISSING
    is_pending: APINullable[bool] = MISSING
    permissions: APINullable[str] = MISSING
    premium_since: APINullable[Optional[Timestamp]] = MISSING
    user: APINullable[User] = MISSING
    avatar: APINullable[str] = MISSING

    @classmethod
    async def from_id(
            cls,
            client: Client,
            guild_id: int,
            _id: int
    ) -> GuildMember:
        data = await client.http.get(f"guilds/{guild_id}/members/{_id}")
        return cls.from_dict({**data, "_client": client, "_http": client.http})



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, IntEnum, auto
from typing import List

from ...utils.api_object import APIObject


@dataclass
class AvailableDevices(APIObject):
    """
    Represents an available device for voice settings

    Attributes
    ----------
    id : :class:`str`
        id of the available device

    name : :class:`str`
        name of the available device
    """

    id: str
    name: str


@dataclass
class VoiceSettingsInput(APIObject):
    """
    Represents a voice setting input object

    Attributes
    ----------
    device_id : :class:`str`
        the device's id

    volume : :class:`float`
        input voice level (min: 0, max: 100)

    available_devices : List[:class:`AvailableDevices`]
        array of read-only device objects containing id and name string keys
    """

    device_id: str
    volume: float
    available_devices: List[AvailableDevices]


@dataclass
class VoiceSettingsOutput(APIObject):
    """
    Represents a voice setting output object

    Attributes
    ----------
    device_id : :class:`str`
        the device's id

    volume : :class:`float`
        input voice level (min: 0, max: 100)

    available_devices : List[:class:`AvailableDevices`]
        array of read-only device objects containing id and name string keys
    """

    device_id: str
    volume: float
    available_devices: List[AvailableDevices]


class VoiceSettingsModeType(Enum):
    """Represents a voice settings mode type"""

    PUSH_TO_TALK = auto()
    VOICE_ACTIVITY = auto()


class KeyTypes(IntEnum):
    """Represents a key type"""

    KEYBOARD_KEY = 0
    MOUSE_BUTTON = 1
    KEYBOARD_MODIFIER_KEY = 2
    GAMEPAD_BUTTON = 3


@dataclass
class ShortcutKeyCombo(APIObject):
    """
    Represents a shortcut key combo for the voice mode settings from a user

    Attributes
    ----------
    type : :class:`KeyTypes`
        type of shortcut key combo

    code : :class:`str`
        key code

    name : :class:`str`
        key name
    """

    type: KeyTypes
    code: int
    name: str


@dataclass
class VoiceSettingsMode(APIObject):
    """
    Represents the voice mode settings from a user

    Attributes
    ----------
    type : :class:`VoiceSettingsModeType`
        voice setting mode type

    auto_threshold : :class:`bool`
        voice activity threshold automatically sets its threshold

    threshold : :class:`float`
        threshold for voice activity (in dB)

    shortcut : :class:`ShortcutKeyCombo`
        shortcut key combos for PTT

    delay : :class:`float`
        the PTT release delay (in ms) (min: 0, max: 2000)
    """

    type: VoiceSettingsModeType
    auto_threshold: bool
    threshold: float
    shortcut: ShortcutKeyCombo
    delay: float


@dataclass
class VoiceSettingsUpdateEvent(APIObject):
    """
    Represents a user's voice settings

    Attributes
    ----------
    input : :class:`VoiceSettingsInput`
        input settings

    output : :class:`VoiceSettingsOutput`
        output settings

    mode : :class:`bool`
        voice mode settings

    automatic_gain_control : :class:`bool`
        state of automatic gain control

    echo_cancellation : :class:`bool`
        state of echo cancellation

    noise_suppression : :class:`bool`
        state of noise suppression

    qos : :class:`bool`
        state of voice quality of service

    silence_warning : :class:`bool`
        state of silence warning notice

    deaf : :class:`bool`
        state of self-deafen

    mute : :class:`bool`
        state of self-mute
    """

    input: VoiceSettingsInput
    output: VoiceSettingsOutput
    automatic_gain_control: bool
    echo_cancellation: bool
    noise_suppression: bool
    qos: bool
    silence_warning: bool
    deaf: bool
    mute: bool



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
non-subscription event sent immediately after connecting,
contains server information
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..commands import ChatCommandHandler
from ..exceptions import InvalidPayload
from ..objects.user.user import User
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import Tuple
    from ..utils.types import Coro
    from ..core.dispatch import GatewayDispatch


async def on_ready_middleware(
    self,
    payload: GatewayDispatch
) -> Tuple[str]:
    """|coro|

    Middleware for ``on_ready`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the stage instance create event

    Returns
    -------
    Tuple[:class:`str`]
        ``on_ready``
    """
    user = payload.data.get("user")
    guilds = payload.data.get("guilds")

    if not user or guilds is None:
        raise InvalidPayload(
            "A `user` and `guilds` key/value pair is expected on the "
            "`ready` payload event."
        )

    self.bot = User.from_dict(construct_client_dict(self, user))
    self.guilds = dict(map(lambda i: (i["id"], None), guilds))

    await ChatCommandHandler(self).initialize()
    return "on_ready",


def export() -> Coro:
    return on_ready_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .app.application import Application
from .app.command import (
    AppCommandType, AppCommandOptionType, AppCommandInteractionDataOption,
    AppCommandOptionChoice, AppCommandOption, AppCommand,
    ClientCommandStructure
)
from .app.intents import Intents
from .app.interaction_base import (
    CallbackType, InteractionType, MessageInteraction
)
from .app.interaction_flags import InteractionFlags
from .app.interactions import (
    ResolvedData, InteractionData, Interaction
)
from .app.select_menu import SelectOption, SelectMenu
from .app.session_start_limit import SessionStartLimit
from .app.throttle_scope import ThrottleScope
from .app.throttling import ThrottleInterface, DefaultThrottleHandler
from .events.channel import ChannelPinsUpdateEvent
from .events.error import DiscordError
from .events.gateway_commands import (
    Identify, Resume, RequestGuildMembers, UpdateVoiceState,
    StatusType, UpdatePresence
)
from .events.guild import (
    GuildBanAddEvent, GuildBanRemoveEvent, GuildEmojisUpdateEvent,
    GuildStickersUpdateEvent, GuildIntegrationsUpdateEvent,
    GuildMemberRemoveEvent, GuildMemberUpdateEvent, GuildMembersChunkEvent,
    GuildRoleCreateEvent, GuildRoleUpdateEvent, GuildRoleDeleteEvent
)
from .events.hello_ready import HelloEvent, ReadyEvent
from .events.integration import IntegrationDeleteEvent
from .events.invite import InviteCreateEvent, InviteDeleteEvent
from .events.message import (
    MessageDeleteEvent, MessageDeleteBulkEvent, MessageReactionAddEvent,
    MessageReactionRemoveEvent, MessageReactionRemoveAllEvent,
    MessageReactionRemoveEmojiEvent
)
from .events.presence import (
    ActivityType, ActivityTimestamp, ActivityEmoji, ActivityParty,
    ActivityAssets, ActivitySecrets, ActivityFlags, ActivityButton,
    Activity, ClientStatus, PresenceUpdateEvent
)
from .events.thread import ThreadListSyncEvent, ThreadMembersUpdateEvent
from .events.typing_start import TypingStartEvent
from .events.voice import VoiceServerUpdateEvent
from .events.webhook import WebhooksUpdateEvent
from .guild.audit_log import (
    AuditLogEvent, AuditLogChange, AuditEntryInfo, AuditLogEntry, AuditLog
)
from .guild.ban import Ban
from .guild.channel import (
    ChannelType, Channel, TextChannel, VoiceChannel, CategoryChannel,
    NewsChannel, ChannelMention
)
from .guild.features import GuildFeature
from .guild.followed_channel import FollowedChannel
from .guild.guild import (
    PremiumTier, GuildNSFWLevel, ExplicitContentFilterLevel, MFALevel,
    VerificationLevel, DefaultMessageNotificationLevel, SystemChannelFlags,
    Guild
)
from .guild.invite import (
    InviteTargetType, InviteStageInstance, InviteMetadata, Invite
)
from .guild.member import GuildMember, PartialGuildMember, BaseMember
from .guild.overwrite import Overwrite
from .guild.role import RoleTags, Role
from .guild.stage import PrivacyLevel, StageInstance
from .guild.template import GuildTemplate
from .guild.thread import ThreadMetadata, ThreadMember
from .guild.webhook import WebhookType, Webhook
from .guild.welcome_screen import WelcomeScreenChannel, WelcomeScreen
from .guild.widget import GuildWidget
from .message.attachment import Attachment
from .message.button import ButtonStyle, Button
from .message.component import MessageComponent
from .message.context import MessageContext
from .message.embed import (
    Embed, EmbedField, EmbedImage, EmbedAuthor, EmbedProvider, EmbedThumbnail,
    EmbedVideo, EmbedFooter
)
from .message.emoji import Emoji
from .message.file import File
from .message.message import Message
from .message.reaction import Reaction
from .message.reference import MessageReference
from .message.sticker import (
    StickerType, StickerFormatType, Sticker, StickerItem, StickerPack
)
from .message.user_message import (
    MessageActivityType, MessageFlags, MessageType, MessageActivity,
    AllowedMentions, AllowedMentionTypes, UserMessage
)
from .user.connection import Connection
from .user.integration import (
    IntegrationExpireBehavior, IntegrationApplication, Integration,
    IntegrationAccount
)
from .user.user import User, PremiumTypes, VisibilityType
from .user.voice_state import VoiceState
from .voice.region import VoiceRegion

__all__ = (
    "Activity", "ActivityAssets", "ActivityButton", "ActivityEmoji",
    "ActivityFlags", "ActivityParty", "ActivitySecrets", "ActivityTimestamp",
    "ActivityType", "AllowedMentionTypes", "AllowedMentions", "AppCommand",
    "AppCommandInteractionDataOption", "AppCommandOption",
    "AppCommandOptionChoice", "AppCommandOptionType", "AppCommandType",
    "Application", "Attachment", "AuditEntryInfo", "AuditLog",
    "AuditLogChange", "AuditLogEntry", "AuditLogEvent", "Ban", "BaseMember",
    "Button", "ButtonStyle", "CallbackType", "CategoryChannel", "Channel",
    "ChannelMention", "ChannelPinsUpdateEvent", "ChannelType",
    "ClientCommandStructure", "ClientStatus", "Connection",
    "DefaultMessageNotificationLevel", "DefaultThrottleHandler",
    "DiscordError", "Embed", "EmbedAuthor", "EmbedField", "EmbedFooter",
    "EmbedImage", "EmbedProvider", "EmbedThumbnail", "EmbedVideo", "Emoji",
    "ExplicitContentFilterLevel", "File", "FollowedChannel", "Guild",
    "GuildBanAddEvent", "GuildBanRemoveEvent", "GuildEmojisUpdateEvent",
    "GuildFeature", "GuildIntegrationsUpdateEvent", "GuildMember",
    "GuildMemberRemoveEvent", "GuildMemberUpdateEvent",
    "GuildMembersChunkEvent", "GuildNSFWLevel", "GuildRoleCreateEvent",
    "GuildRoleDeleteEvent", "GuildRoleUpdateEvent", "GuildStickersUpdateEvent",
    "GuildTemplate", "GuildWidget", "HelloEvent", "Identify", "Integration",
    "IntegrationAccount", "IntegrationApplication", "IntegrationDeleteEvent",
    "IntegrationExpireBehavior", "Intents", "Interaction", "InteractionData",
    "InteractionFlags", "InteractionType", "Invite", "InviteCreateEvent",
    "InviteDeleteEvent", "InviteMetadata", "InviteStageInstance",
    "InviteTargetType", "MFALevel", "Message", "MessageActivity",
    "MessageActivityType", "MessageComponent", "MessageContext",
    "MessageDeleteBulkEvent", "MessageDeleteEvent", "MessageFlags",
    "MessageInteraction", "MessageReactionAddEvent",
    "MessageReactionRemoveAllEvent", "MessageReactionRemoveEmojiEvent",
    "MessageReactionRemoveEvent", "MessageReference", "MessageType",
    "NewsChannel", "Overwrite", "PartialGuildMember", "PremiumTier",
    "PremiumTypes", "PresenceUpdateEvent", "PrivacyLevel", "Reaction",
    "ReadyEvent", "RequestGuildMembers", "ResolvedData", "Resume", "Role",
    "RoleTags", "SelectMenu", "SelectOption", "SessionStartLimit",
    "StageInstance", "StatusType", "Sticker", "StickerFormatType",
    "StickerItem", "StickerPack", "StickerType", "SystemChannelFlags",
    "TextChannel", "ThreadListSyncEvent", "ThreadMember",
    "ThreadMembersUpdateEvent", "ThreadMetadata", "ThrottleInterface",
    "ThrottleScope", "TypingStartEvent", "UpdatePresence", "UpdateVoiceState",
    "User", "UserMessage", "VerificationLevel", "VisibilityType",
    "VoiceChannel", "VoiceRegion", "VoiceServerUpdateEvent", "VoiceState",
    "Webhook", "WebhookType", "WebhooksUpdateEvent", "WelcomeScreen",
    "WelcomeScreenChannel"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
Sent in response to Guild Request Members. You can use the ``chunk_index``
and ``chunk_count`` to calculate how many chunks are left for your request.
"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.guild import GuildMembersChunkEvent
from ..utils import Coro
from ..utils.conversion import construct_client_dict


async def guild_member_chunk_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_member_chunk`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild member chunk event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.guild.GuildMemberChunkEvent`]]
        ``on_guild_member_chunk`` and a ``GuildMemberChunkEvent``
    """

    return (
        "on_guild_member_chunk",
        [GuildMembersChunkEvent.from_dict(
            construct_client_dict(self, payload.data)
        )]
    )


def export() -> Coro:
    return guild_member_chunk_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import APINullable, MISSING

if TYPE_CHECKING:
    from ..user.user import User
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp
    from ..guild.invite import InviteTargetType


@dataclass
class InviteCreateEvent(APIObject):
    """Sent when a new invite to a channel is created.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The channel the invite is for
    code: :class:`str`
        The unique invite code
    created_at: :class:`~pincer.utils.timestamp.Timestamp`
        The time at which the invite was created
    max_age: :class:`int`
        How long the invite is valid for (in seconds)
    max_uses: :class:`int`
        The maximum number of times the invite can be used
    temporary: :class:`bool`
        Whether or not the invite is temporary (invited users will
        be kicked on disconnect unless they're assigned a role)
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The guild of the invite
    inviter: APIObject[:class:`~pincer.objects.user.user.User`]
        The user that created the invite
    target_type: APINullable[:class:`~pincer.objects.guild.invite.InviteTargetType`]
        The type of target for this voice channel invite
    target_user: APINullable[:class:`~pincer.objects.user.user.User`]
        The user whose stream to display for
        this voice channel stream invite
    uses: :class:`int`
        How many times the invite has been used (always will be ``0``)
    """
    # noqa: E501
    channel_id: Snowflake
    code: str
    created_at: Timestamp
    max_age: int
    max_uses: int
    temporary: bool

    guild_id: APINullable[Snowflake] = MISSING
    inviter: APINullable[User] = MISSING
    target_type: APINullable[InviteTargetType] = MISSING
    target_user: APINullable[User] = MISSING
    uses: int = 0


@dataclass
class InviteDeleteEvent(APIObject):
    """Sent when an invite is deleted.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        The channel of the invite
    code: :class:`str`
        The unique invite code
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The guild of the invite
    """
    channel_id: Snowflake
    code: str

    guild_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .audit_log import (
    AuditLogEvent, AuditLogChange, AuditEntryInfo, AuditLogEntry, AuditLog
)
from .ban import Ban
from .channel import (
    ChannelType, Channel, TextChannel, VoiceChannel, CategoryChannel,
    NewsChannel, ChannelMention
)
from .features import GuildFeature
from .followed_channel import FollowedChannel
from .guild import (
    PremiumTier, GuildNSFWLevel, ExplicitContentFilterLevel, MFALevel,
    VerificationLevel, DefaultMessageNotificationLevel, SystemChannelFlags,
    Guild, UnavailableGuild
)
from .invite import (
    InviteTargetType, InviteStageInstance, InviteMetadata, Invite
)
from .member import GuildMember, PartialGuildMember, BaseMember
from .overwrite import Overwrite
from .role import RoleTags, Role
from .stage import PrivacyLevel, StageInstance
from .template import GuildTemplate
from .thread import ThreadMetadata, ThreadMember
from .webhook import WebhookType, Webhook
from .welcome_screen import WelcomeScreenChannel, WelcomeScreen
from .widget import GuildWidget


__all__ = (
    "AuditEntryInfo", "AuditLog", "AuditLogChange", "AuditLogEntry",
    "AuditLogEvent", "Ban", "BaseMember", "CategoryChannel", "Channel",
    "ChannelMention", "ChannelType", "DefaultMessageNotificationLevel",
    "ExplicitContentFilterLevel", "FollowedChannel", "Guild", "GuildFeature",
    "GuildMember", "GuildNSFWLevel", "GuildTemplate", "GuildWidget", "Invite",
    "InviteMetadata", "InviteStageInstance", "InviteTargetType", "MFALevel",
    "NewsChannel", "Overwrite", "PartialGuildMember", "PremiumTier",
    "PrivacyLevel", "Role", "RoleTags", "StageInstance", "SystemChannelFlags",
    "TextChannel", "ThreadMember", "ThreadMetadata", "UnavailableGuild",
    "VerificationLevel", "VoiceChannel", "Webhook", "WebhookType",
    "WelcomeScreen", "WelcomeScreenChannel"
)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import io
from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from aiohttp import ClientSession

from ...utils.api_object import APIObject
from ...utils.conversion import construct_client_dict
from ...utils.types import MISSING

PILLOW_IMPORT = True

try:
    from PIL import Image
except (ModuleNotFoundError, ImportError):
    PILLOW_IMPORT = False


if TYPE_CHECKING:
    from typing import Optional

    from ...client import Client
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


class PremiumTypes(IntEnum):
    """The type of Discord premium a user has.

    Attributes
    ----------
    NONE:
        No nitro.
    NITRO_CLASSIC:
        Nitro classic (no boosts).
    NITRO:
        Full nitro subscription.
    """

    NONE = 0
    NITRO_CLASSIC = 1
    NITRO = 2


class VisibilityType(IntEnum):
    """The type of a connection visibility.

    Attributes
    ----------
    NONE:
        Connection is not visible.
    EVERYONE:
        Connection is visible to everyone.
    """

    NONE = 0
    EVERYONE = 1


@dataclass
class User(APIObject):
    """Represents a Discord user. This can be a bot account or a
    human account.

    Attributes
    ----------
    avatar: Optional[:class:`str`]
        The user's avatar hash
    discriminator: :class:`str`
        The user's 4-digit discord-tag
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The user's id
    username: :class:`str`
        The user's username, not unique across the platform
    flags: APINullable[:class:`int`]
        The flags on a user's account
    accent_color: APINullable[Optional[:class:`int`]]
        The user's banner color encoded as an integer representation of
        hexadecimal color code
    banner: APINullable[Optional[:class:`str`]]
        The user's banner, or null if unset
    banner_color: APINullable[Optional[:class:`int`]]
        The color of the user's banner
    bot: APINullable[:class:`bool`]
        Whether the user belongs to an OAuth2 application
    email: APINullable[Optional[:class:`str`]]
        The user's email
    locale: APINullable[:class:`str`]
        The user's chosen language option
    mfa_enabled: APINullable[:class:`bool`]
        Whether the user has two factor enabled on their account
    premium_type: APINullable[:class:`int`]
        The type of Nitro subscription on a user's account
    public_flags: APINullable[:class:`int`]
        The public flags on a user's account
    system: APINullable[:class:`bool`]
        Whether the user is an Official Discord System user
        (part of the urgent message system)
    verified: APINullable[:class:`bool`]
        Whether the email on this account has been verified
    """

    id: Snowflake
    username: APINullable[str] = MISSING
    discriminator: APINullable[str] = MISSING

    avatar: APINullable[str] = MISSING
    flags: APINullable[int] = MISSING
    accent_color: APINullable[Optional[int]] = MISSING
    banner: APINullable[Optional[str]] = MISSING
    banner_color: APINullable[Optional[int]] = MISSING
    bot: APINullable[bool] = MISSING
    email: APINullable[Optional[str]] = MISSING
    locale: APINullable[str] = MISSING
    mfa_enabled: APINullable[bool] = MISSING
    premium_type: APINullable[int] = MISSING
    public_flags: APINullable[int] = MISSING
    system: APINullable[bool] = MISSING
    verified: APINullable[bool] = MISSING

    @property
    def premium(self) -> APINullable[PremiumTypes]:
        """APINullable[:class:`~pincer.objects.user.user.PremiumTypes`]: The
        user their premium type in a usable enum.
        """
        return (
            MISSING if self.premium_type is MISSING else PremiumTypes(self.premium_type)
        )

    @property
    def mention(self) -> str:
        """:class:`str`: The user's mention string."""
        return f"<@!{self.id}>"

    def get_avatar_url(self, size: int = 512, ext: str = "png") -> str:
        return (
            f"https://cdn.discordapp.com/avatars/{self.id}/{self.avatar}.{ext}"
            f"?size={size}"
        )

    if PILLOW_IMPORT:

        async def get_avatar(self, size: int = 512, ext: str = "png") -> Image:
            """Get the user's avatar as a Pillow image.

            Parameters
            ----------
            size : :class: int, optional
                The size of the image to get. Defaults to 512.
            ext : :class: str, optional
                The file extension to use. Defaults to 'png'.

            Returns
            -------
            :class: Image
                The user's avatar as a Pillow image.
            """
            async with ClientSession().get(url=self.get_avatar_url()) as resp:
                avatar = io.BytesIO(await resp.read())
                print(Image, dir(Image))
                return Image.open(avatar).convert("RGBA")

    def __str__(self):
        return self.username + "#" + self.discriminator

    @classmethod
    async def from_id(cls, client: Client, user_id: int) -> User:
        data = await client.http.get(f"users/{user_id}")
        return cls.from_dict(construct_client_dict(client, data))



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import APINullable, MISSING

if TYPE_CHECKING:
    from ..guild.member import GuildMember
    from ...utils.snowflake import Snowflake


@dataclass
class TypingStartEvent(APIObject):
    """Sent when a user starts typing in a channel.

    Attributes
    ----------
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the channel
    user_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the user
    timestamp: :class:`int`
        Unix time (in seconds) of when the user started typing
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the guild
    member: APINullable[:class:`~pincer.objects.guild.member.GuildMember`]
        The member who started typing if this happened in a guild
    """
    channel_id: Snowflake
    user_id: Snowflake
    timestamp: int

    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING



from __future__ import annotations

from contextlib import contextmanager
from os import getcwd, chdir as os_chdir
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Generator, Any


@contextmanager
def chdir(path: str) -> Generator[str, Any, None]:
    """Change the cwd to the given path as a context manager.

    Parameters
    ----------
    path:
        Path to change to.

    Yields
    ------
    Generator[:class:`str`, Any, None]
    """
    current_path = getcwd()
    os_chdir(path)

    yield current_path
    os_chdir(current_path)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import Event, wait_for as _wait_for, get_running_loop, TimeoutError
from collections import deque
from typing import TYPE_CHECKING

from ..exceptions import TimeoutError as PincerTimeoutError

if TYPE_CHECKING:
    from typing import Any, List, Union, Optional
    from .types import CheckFunction


class _Processable(ABC):

    @abstractmethod
    def process(self, event_name: str, *args):
        """
        Method that is ran when an event is received from discord.

        Parameters
        ----------
        event_name : str
            The name of the event.
        *args : Any
            Arguments to evaluate check with.

        Returns
        -------
        bool
            Whether the event can be set
        """

    def matches_event(self, event_name: str, *args):
        """
        Parameters
        ----------
        event_name : str
            Name of event.
        *args : Any
            Arguments to eval check with.
        """
        if self.event_name != event_name:
            return False

        if self.check:
            return self.check(*args)

        return True


def _lowest_value(*args):
    """
    Returns lowest value from list of numbers. ``None`` is not counted as a
    value. ``None`` is returned if all arguments are ``None``.
    """
    args_without_none = [n for n in args if n is not None]

    if not args_without_none:
        return None

    return min(args_without_none)


class _Event(_Processable):
    """
    Parameters
    ----------
    event_name : str
        The name of the event.
    check : Optional[Callable[[Any], bool]]
        ``can_be_set`` only returns true if this function returns true.
        Will be ignored if set to None.

    Attributes
    ----------
    event : :class:`asyncio.Event`
        Even that is used to wait until the next valid discord event.
    return_value : Optional[str]
        Used to store the arguments from ``can_be_set`` so they can be
        returned later.
    """

    def __init__(
        self,
        event_name: str,
        check: CheckFunction
    ):
        self.event_name = event_name
        self.check = check
        self.event = Event()
        self.return_value = None
        super().__init__()

    async def wait(self):
        """
        Waits until ``self.event`` is set.
        """
        await self.event.wait()

    def process(self, event_name: str, *args) -> bool:
        if self.matches_event(event_name, *args):
            self.return_value = args
            self.event.set()


class _LoopEmptyError(Exception):
    """Raised when the _LoopMgr is empty and cannot accept new item"""


class _LoopMgr(_Processable):
    """
    Parameters
    ----------
    event_name : str
        The name of the event.
    check : Optional[Callable[[Any], bool]]
        ``can_be_set`` only returns true if this function returns true.
        Will be ignored if set to None.

    Attributes
    ----------
    can_expand : bool
        Whether the queue is allowed to grow. Turned to false once the
        EventMgr's timer runs out.
    events : :class:`collections.deque`
        Queue of events to be processed.
    wait : :class:`asyncio.Event`
        Used to make ``get_next()` wait for the next event.
    """

    def __init__(self, event_name: str, check: CheckFunction) -> None:
        self.event_name = event_name
        self.check = check

        self.can_expand = True
        self.events = deque()
        self.wait = Event()

    def process(self, event_name: str, *args):
        if not self.can_expand:
            return

        if self.matches_event(event_name, *args):
            self.events.append(args)
            self.wait.set()

    async def get_next(self):
        """
        Returns the next item if the queue. If there are no items in the queue,
        it will return the next event that happens.
        """
        if not self.events:
            if not self.can_expand:
                raise _LoopEmptyError

            self.wait.clear()
            await self.wait.wait()
        return self.events.popleft()


class EventMgr:
    """
    Attributes
    ----------
    event_list : List[_DiscordEvent]
        The List of events that need to be processed.
    """

    def __init__(self):
        self.event_list: List[_Processable] = []

    def process_events(self, event_name, *args):
        """
        Parameters
        ----------
        event_name : str
            The name of the event to be processed.
        *args : Any
            The arguments returned from the middleware for this event.
        """
        for event in self.event_list:
            event.process(event_name, *args)

    async def wait_for(
        self,
        event_name: str,
        check: CheckFunction,
        timeout: Optional[float]
    ) -> Any:
        """
        Parameters
        ----------
        event_name : str
            The type of event. It should start with `on_`. This is the same
            name that is used for @Client.event.
        check : Union[Callable[[Any], bool], None]
            This function only returns a value if this return true.
        timeout: Union[float, None]
            Amount of seconds before timeout. Use None for no timeout.

        Returns
        ------
        Any
            What the Discord API returns for this event.
        """

        event = _Event(event_name, check)
        self.event_list.append(event)

        try:
            await _wait_for(event.wait(), timeout=timeout)
        except TimeoutError:
            raise PincerTimeoutError(
                "wait_for() timed out while waiting for an event."
            )
        self.event_list.remove(event)
        return event.return_value

    async def loop_for(
        self,
        event_name: str,
        check: CheckFunction,
        iteration_timeout: Optional[float],
        loop_timeout: Optional[float],
    ) -> Any:
        """
        Parameters
        ----------
        event_name : str
            The type of event. It should start with `on_`. This is the same
            name that is used for @Client.event.
        check : Callable[[Any], bool]
            This function only returns a value if this return true.
        iteration_timeout: Union[float, None]
            Amount of seconds before timeout. Timeouts are for each loop.
        loop_timeout: Union[float, None]
            Amount of seconds before the entire loop times out. The generator
            will only raise a timeout error while it is waiting for an event.

        Yields
        ------
        Any
            What the Discord API returns for this event.
        """

        loop_mgr = _LoopMgr(event_name, check)
        self.event_list.append(loop_mgr)

        loop = get_running_loop()

        while True:
            start_time = loop.time()

            try:
                yield await _wait_for(
                    loop_mgr.get_next(),
                    timeout=_lowest_value(
                        loop_timeout, iteration_timeout
                    )
                )

            except TimeoutError:
                # Loop timed out. Loop through the remaining events received
                # before the timeout.
                loop_mgr.can_expand = False
                try:
                    while True:
                        yield await loop_mgr.get_next()
                except _LoopEmptyError:
                    raise PincerTimeoutError(
                        "loop_for() timed out while waiting for an event"
                    )

            # `not` can't be used here because there is a check for
            # `loop_timeout == 0`
            if loop_timeout is not None:
                loop_timeout -= loop.time() - start_time

                # loop_timeout can be below 0 if the user's code in the for loop
                # takes longer than the time left in loop_timeout
                if loop_timeout <= 0:
                    raise PincerTimeoutError(
                        "loop_for() timed out while waiting for an event"
                    )



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List, Optional

    from ..user import User
    from ...utils import APINullable, Snowflake


class StickerType(IntEnum):
    """Displays from where the sticker comes from.

    Attributes
    ----------
    STANDARD:
        Sticker is included in the default Discord sticker pack.
    GUILD:
        Sticker is a custom sticker from a discord server.
    """
    STANDARD = 1
    GUILD = 2


class StickerFormatType(IntEnum):
    """The type of the sticker.

    Attributes
    ----------
    PNG:
        Sticker is of PNG format.
    APNG:
        Sticker is animated with APNG format.
    LOTTIE:
        Sticker is animated with with LOTTIE format. (vector based)
    """
    PNG = 1
    APNG = 2
    LOTTIE = 3


@dataclass
class Sticker(APIObject):
    """Represents a Discord sticker.

    Attributes
    ----------
    description: Optional[:class:`str`]
        description of the sticker
    format_type: :class:`~pincer.objects.message.sticker.StickerFormatType`
        type of sticker format
    id: :class:`~pincer.utils.snowflake.Snowflake`
        id of the sticker
    name: :class:`str`
        name of the sticker
    tags: :class:`str`
        for guild stickers, the Discord name of a unicode emoji
        representing the sticker's expression. For standard stickers,
        a comma-separated list of related expressions.
    type: :class:`~pincer.objects.message.sticker.StickerType`
        type of sticker
    available: APINullable[:class:`bool`]
        whether this guild sticker can be used,
        may be false due to loss of Server Boosts
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        id of the guild that owns this sticker
    pack_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        for standard stickers, id of the pack the sticker is from
    sort_value: APINullable[:class:`int`]
        the standard sticker's sort order within its pack
    user: APINullable[:class:`~pincer.objects.user.user.User`]
        the user that uploaded the guild sticker
    """

    description: Optional[str]
    format_type: StickerFormatType
    id: Snowflake
    name: str
    tags: str
    type: StickerType

    available: APINullable[bool] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    pack_id: APINullable[Snowflake] = MISSING
    sort_value: APINullable[int] = MISSING
    user: APINullable[User] = MISSING


@dataclass
class StickerItem(APIObject):
    """Represents the smallest amount of data required to render a sticker.
    A partial sticker object.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the sticker
    name: :class:`str`
        Name of the sticker
    format_type: :class:`~pincer.objects.message.sticker.StickerFormatType`
        Type of sticker format
    """

    id: Snowflake
    name: str
    format_type: StickerFormatType


@dataclass
class StickerPack(APIObject):
    """Represents a pack of standard stickers.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the sticker pack
    stickers: List[:class:`~pincer.objects.message.sticker.Sticker`]
        The stickers in the pack
    name: :class:`str`
        Name of the sticker pack
    sku_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the pack's SKU
    description: :class:`str`
        Description of the sticker pack
    cover_sticker_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of a sticker in the pack which is shown as the pack's icon
    banner_asset_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the sticker pack's banner image
    """

    id: Snowflake
    stickers: List[Sticker]
    name: str
    sku_id: Snowflake
    description: str

    cover_sticker_id: APINullable[Snowflake] = MISSING
    banner_asset_id: APINullable[Snowflake] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List

    from ..message.emoji import Emoji
    from ...utils.types import APINullable


@dataclass
class SelectOption(APIObject):
    """Represents a Discord Select Option object

    Attributes
    ----------
    label: :class:`str`
        The user-facing name of the option, max 100 characters
    value: :class:`str`
        The def-defined value of the option, max 100 characters
    description: APINullable[:class:`str`]
        An additional description of the option, max 100 characters
    emoji: APINullable[:class:`~pincer.objects.message.emoji.Emoji`]
        ``id``, ``name``, and ``animated``
    default: APINullable[:class:`bool`]
        Will render this option as selected by default
    """
    label: str
    value: str
    description: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    default: APINullable[bool] = MISSING


@dataclass
class SelectMenu(APIObject):
    """Represents a Discord Select Menu object

    Attributes
    ----------
    type: :class:`int`
        ``3`` for a select menu
    custom_id: :class:`str`
        A developer-defined identifier for the button,
        max 100 characters
    options: List[:class:`~pincer.objects.app.select_menu.SelectOption`]
        The choices in the select, max ``25``
    placeholder: APINullable[:class:`str`]
        Custom placeholder text if nothing is selected,
        max 100 characters
    min_values: APINullable[:class:`int`]
        The minimum number of items that must be chosen; min ``0``, max ``25``
        |default| ``1``
    max_values: APINullable[:class:`int`]
        The maximum number of items that can be chosen; max 25
        |default| ``1``
    disabled: APINullable[:class:`bool`]
        Disable the select
        |default| False
    """
    type: int
    custom_id: str
    options: List[SelectOption]

    placeholder: APINullable[str] = MISSING
    min_values: APINullable[int] = 1
    max_values: APINullable[int] = 1
    disabled: APINullable[bool] = False



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import List

    from ..app.select_menu import SelectOption
    from ..message.button import ButtonStyle
    from ..message.emoji import Emoji
    from ...utils.types import APINullable


@dataclass
class MessageComponent(APIObject):
    """Represents a Discord Message Component object

    Attributes
    ----------
    type: :class:`int`
        Component type
    options: List[:class:`~pincer.objects.app.select_menu.SelectOption`]
        The choices in the select, max 25
    custom_id: APINullable[:class:`str`]
        A developer-defined identifier for the component,
        max 100 characters
    disabled: APINullable[:class:`bool`]
        Whether the component is disabled,
        defaults to `False`
    style: APINullable[:class:`~pincer.objects.message.button.ButtonStyle`]
        One of button styles
    label: APINullable[:class:`str`]
        Text that appears on the button, max 80 characters
    emoji: APINullable[:class:`~pincer.objects.message.emoji.Emoji`]
        ``name``, ``id``, and ``animated``
    url: APINullable[:class:`str`]
        A url for link-style buttons
    placeholder: APINullable[:class:`str`]
        Custom placeholder text if nothing is selected,
        max 100 characters
    min_values: APINullable[:class:`int`]
        The minimum number of items that must be chosen;
        |default| ``1``, min ``0``, max ``25``
    max_values: APINullable[:class:`int`]
        The maximum number of items that can be chosen;
        |default| ``1``, max ``25``
    components: APINullable[List[:class:`~pincer.objects.message.component.MessageComponent`]]
        A list of child components
    """
    # noqa: E501
    type: int
    options: List[SelectOption]

    custom_id: APINullable[str] = MISSING
    disabled: APINullable[bool] = False
    style: APINullable[ButtonStyle] = MISSING
    label: APINullable[str] = MISSING
    emoji: APINullable[Emoji] = MISSING
    url: APINullable[str] = MISSING
    placeholder: APINullable[str] = MISSING
    min_values: APINullable[int] = 1
    max_values: APINullable[int] = 1
    components: APINullable[List[MessageComponent]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from inspect import getfullargspec, Parameter, Signature
from typing import Any, Union, Callable, Mapping, List

from .types import Coro, TypeCache
from ..objects.message.context import MessageContext


def should_pass_cls(call: Union[Coro, Callable[..., Any]]) -> bool:
    """
    Checks whether a callable requires a self/cls as first parameter.

    :param call:
        The callable to check.

    :return:
        Whether or not its required.
    """
    args = getfullargspec(call).args
    return len(args) >= 1 and args[0] in ["self", "cls"]


context_types = [Signature.empty, MessageContext]


def should_pass_ctx(sig: Mapping[str, Parameter], params: List[str]) -> bool:
    # TODO: Write docs, do not put online
    if not params:
        return False

    annotation = sig[params[0]].annotation
    if isinstance(annotation, str):
        TypeCache()
        annotation = eval(annotation, TypeCache.cache, globals())

    return annotation in context_types



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""
non-subscription event sent when there is an error,
including command responses
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.events.error import DiscordError
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro

if TYPE_CHECKING:
    from typing import List, Tuple
    from ..core.dispatch import GatewayDispatch


def error_middleware(
    self,
    payload: GatewayDispatch
) -> Tuple[str, List[DiscordError]]:
    """|coro|

    Middleware for ``on_error`` event.

    Parameters
    ----------
    payload : :class:`~pincer.core.dispatch.GatewayDispatch`
        The data received from the ready event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.events.error.DiscordError`]]
        ``"on_error"`` and a ``DiscordError``
    """
    # noqa: E501

    return "on_error", [
        DiscordError.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return error_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a guild is created/joined on the client"""
from __future__ import annotations

from typing import TYPE_CHECKING

from ..objects.guild import Guild
from ..utils.conversion import construct_client_dict

if TYPE_CHECKING:
    from typing import List, Tuple
    from ..core.dispatch import GatewayDispatch


async def guild_create_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_guild_create``,
        generate the guild class that was created

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the guild create event

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.Guild`]]

        ``on_guild_create`` and a ``Guild``
    """
    guild = Guild.from_dict(construct_client_dict(self, payload.data))
    self.guilds[guild.id] = guild
    return "on_guild_create", [guild]


def export():
    return guild_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from asyncio import sleep, ensure_future
from dataclasses import dataclass
from enum import IntEnum
from typing import overload, TYPE_CHECKING

from ..message.user_message import UserMessage
from ..._config import GatewayConfig
from ...utils.api_object import APIObject
from ...utils.conversion import construct_client_dict
from ...utils.convert_message import convert_message
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Union

    from .member import GuildMember
    from .overwrite import Overwrite
    from .thread import ThreadMetadata
    from ..message.message import Message
    from ..message.embed import Embed
    from ..user.user import User
    from ...client import Client
    from ...utils.timestamp import Timestamp
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


class ChannelType(IntEnum):
    """Represents a channel its type.

    Attributes
    ----------
    GUILD_TEXT:
        A text channel.
    DM:
        A DM channel.
    GUILD_VOICE:
        A voice channel.
    GROUP_DM:
        A group DM channel.
    GUILD_CATEGORY:
        A category channel.
    GUILD_NEWS:
        A news channel.
    GUILD_STORE:
        A store channel.
    GUILD_NEWS_THREAD:
        A news thread.
    GUILD_PUBLIC_THREAD:
        A public thread.
    GUILD_PRIVATE_THREAD:
        A private thread.
    GUILD_STAGE_VOICE:
        A stage channel.
    """
    GUILD_TEXT = 0
    DM = 1
    GUILD_VOICE = 2
    GROUP_DM = 3
    GUILD_CATEGORY = 4
    GUILD_NEWS = 5
    GUILD_STORE = 6

    if GatewayConfig.version >= 9:
        GUILD_NEWS_THREAD = 10
        GUILD_PUBLIC_THREAD = 11
        GUILD_PRIVATE_THREAD = 12

    GUILD_STAGE_VOICE = 13


@dataclass
class Channel(APIObject):  # noqa E501
    """Represents a Discord Channel Mention object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The id of this channel
    type: :class:`~pincer.objects.guild.channel.ChannelType`
        The type of channel
    application_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Application id of the group DM creator if it is bot-created
    bitrate: APINullable[:class:`int`]
        The bitrate (in bits) of the voice channel
    default_auto_archive_duration: APINullable[:class:`int`]
        Default duration for newly created threads, in minutes, to
        automatically archive the thread after recent activity, can be set to:
        60, 1440, 4320, 10080
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the guild (may be missing for some channel objects received
        over gateway guild dispatches)
    icon: APINullable[Optional[:class:`str`]]
        Icon hash
    last_message_id: APINullable[Optional[:class:`~pincer.utils.snowflake.Snowflake`]]
        The id of the last message sent in this channel (may not point to an
        existing or valid message)
    last_pin_timestamp: APINullable[Optional[:class:`~pincer.utils.timestamp.Timestamp`]]
        When the last pinned message was pinned. This may be null in events
        such as GUILD_CREATE when a message is not pinned.
    member: APINullable[:class:`~pincer.objects.guild.member.GuildMember`]
        Thread member object for the current user, if they have joined the
        thread, only included on certain API endpoints
    member_count: APINullable[:class:`int`]
        An approximate count of users in a thread, stops counting at 50
    message_count: :class:`int`
        An approximate count of messages in a thread, stops counting at 50
    name: APINullable[:class:`str`]
        The name of the channel (1-100 characters)
    nsfw: APINullable[:class:`bool`]
        Whether the channel is nsfw
    owner_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        Id of the creator of the group DM or thread
    parent_id: APINullable[Optional[:class:`~pincer.utils.snowflake.Snowflake`]]
        For guild channels: id of the parent category for a channel (each
        parent category can contain up to 50 channels), for threads: id of the
        text channel this thread was created
    permissions: APINullable[:class:`str`]
        Computed permissions for the invoking user in the channel, including
        overwrites, only included when part of the resolved data received on a
        slash command interaction
    permission_overwrites: APINullable[List[:class:`~pincer.objects.guild.overwrite.Overwrite`]]
        Explicit permission overwrites for members and roles
    position: APINullable[:class:`int`]
        Sorting position of the channel
    rate_limit_per_user: APINullable[:class:`int`]
        Amount of seconds a user has to wait before sending another message
        (0-21600); bots, as well as users with the permission manage_messages
        or manage_channel, are unaffected
    recipients: APINullable[List[:class:`~pincer.objects.user.user.User`]]
        The recipients of the DM
    rtc_region: APINullable[Optional[:class:`str`]]
        Voice region id for the voice channel, automatic when set to null
    thread_metadata: APINullable[:class:`~pincer.objects.guild.thread.ThreadMetadata`]
        Thread-specific fields not needed by other channels
    topic: APINullable[Optional[:class:`str`]]
        The channel topic (0-1024 characters)
    user_limit: APINullable[:class:`int`]
        The user limit of the voice channel
    video_quality_mode: APINullable[:class:`int`]
        The camera video quality mode of the voice channel, 1 when not present
    """
    # noqa: E501
    id: Snowflake
    type: ChannelType

    application_id: APINullable[Snowflake] = MISSING
    bitrate: APINullable[int] = MISSING
    default_auto_archive_duration: APINullable[int] = MISSING
    guild_id: APINullable[Snowflake] = MISSING
    icon: APINullable[Optional[str]] = MISSING

    last_message_id: APINullable[Optional[Snowflake]] = MISSING
    last_pin_timestamp: APINullable[Optional[Timestamp]] = MISSING
    member: APINullable[GuildMember] = MISSING

    member_count: APINullable[int] = MISSING

    message_count: APINullable[int] = MISSING

    name: APINullable[str] = MISSING
    nsfw: APINullable[bool] = MISSING
    owner_id: APINullable[Snowflake] = MISSING
    parent_id: APINullable[Optional[Snowflake]] = MISSING
    permissions: APINullable[str] = MISSING
    permission_overwrites: APINullable[List[Overwrite]] = MISSING
    position: APINullable[int] = MISSING
    rate_limit_per_user: APINullable[int] = MISSING
    recipients: APINullable[List[User]] = MISSING
    rtc_region: APINullable[Optional[str]] = MISSING
    thread_metadata: APINullable[ThreadMetadata] = MISSING
    topic: APINullable[Optional[str]] = MISSING
    user_limit: APINullable[int] = MISSING
    video_quality_mode: APINullable[int] = MISSING

    @property
    def mention(self):
        return f"<#{self.id}>"

    @classmethod
    async def from_id(cls, client: Client, channel_id: int) -> Channel:
        # TODO: Write docs
        data = (await client.http.get(f"channels/{channel_id}")) or {}

        data.update(construct_client_dict(
            client,
            {"type": ChannelType(data.pop("type"))}
        ))

        channel_cls = _channel_type_map.get(data["type"], Channel)
        return channel_cls.from_dict(data)

    @overload
    async def edit(
            self, *, name: str = None,
            type: ChannelType = None,
            position: int = None, topic: str = None, nsfw: bool = None,
            rate_limit_per_user: int = None, bitrate: int = None,
            user_limit: int = None,
            permissions_overwrites: List[Overwrite] = None,
            parent_id: Snowflake = None, rtc_region: str = None,
            video_quality_mod: int = None,
            default_auto_archive_duration: int = None
    ) -> Channel:
        ...

    async def edit(
            self,
            reason: Optional[str] = None,
            **kwargs
    ):
        """Edit a channel with the given keyword arguments.

        Parameters
        ----------
        reason Optional[:class:`str`]
            The reason of the channel delete.
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        headers = {}

        if reason is not None:
            headers["X-Audit-Log-Reason"] = str(reason)

        data = await self._http.patch(
            f"channels/{self.id}",
            kwargs,
            headers=headers
        )
        data.update(construct_client_dict(
            self._client,
            {"type": ChannelType(data.pop("type"))}
        ))
        channel_cls = _channel_type_map.get(data["type"], Channel)
        return channel_cls.from_dict(data)

    async def delete(
            self,
            reason: Optional[str] = None,
            /,
            channel_id: Optional[Snowflake] = None
    ):
        """|coro|

        Delete the current channel.

        Parameters
        ----------
        reason Optional[:class:`str`]
            The reason of the channel delete.
        channel_id :class:`~.pincer.utils.Snowflake`
            The id of the channel, defaults to the current object id.
        """
        channel_id = channel_id or self.id

        headers = {}

        if reason is not None:
            headers["X-Audit-Log-Reason"] = str(reason)

        await self._http.delete(
            f"channels/{channel_id}",
            headers
        )

    async def __post_send_handler(self, message: Message):
        """Process a message after it was sent.

        Parameters
        ----------
        message :class:`~.pincer.objects.message.message.Message`
            The message.
        """

        if getattr(message, "delete_after", None):
            await sleep(message.delete_after)
            await self.delete()

    def __post_sent(
            self,
            message: Message
    ):
        """Ensure the `__post_send_handler` method its future.

        Parameters
        ----------
        message :class:`~.pincer.objects.message.message.Message`
            The message.
        """
        ensure_future(self.__post_send_handler(message))

    async def send(self, message: Union[Embed, Message, str]) -> UserMessage:
        """|coro|

        Send a message in the channel.

        Parameters
        ----------
        message :class:`~.pincer.objects.message.message.Message`
            The message which must be sent

        Returns
        -------
        :class:`~.pincer.objects.message.user_message.UserMessage`
            The message that was sent.
        """
        content_type, data = convert_message(self._client, message).serialize()

        resp = await self._http.post(
            f"channels/{self.id}/messages",
            data,
            content_type=content_type
        )
        msg = UserMessage.from_dict(resp)
        self.__post_sent(msg)
        return msg

    def __str__(self):
        """return the discord tag when object gets used as a string."""
        return self.name or str(self.id)


class TextChannel(Channel):
    """A subclass of ``Channel`` for text channels with all the same attributes."""

    @overload
    async def edit(
            self, name: str = None, type: ChannelType = None,
            position: int = None, topic: str = None, nsfw: bool = None,
            rate_limit_per_user: int = None,
            permissions_overwrites: List[Overwrite] = None,
            parent_id: Snowflake = None,
            default_auto_archive_duration: int = None
    ) -> Union[TextChannel, NewsChannel]:
        ...

    async def edit(self, **kwargs):
        """Edit a text channel with the given keyword arguments.

        Parameters
        ----------
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        return await super().edit(**kwargs)

    async def fetch_message(self, message_id: int) -> UserMessage:
        """|coro|
        Returns a UserMessage from this channel with the given id.

        Parameters
        ----------
        message_id : :class: int
            The message ID to look for.

        Returns
        -------
        :class:`~pincer.objects.message.user_message.UserMessage`
            The requested message.
        """
        return UserMessage.from_dict(
            await self._http.get(
                f"/channels/{self.id}/messages/{message_id}"
            )
        )


class VoiceChannel(Channel):
    """A subclass of ``Channel`` for voice channels with all the same attributes."""

    @overload
    async def edit(
            self, name: str = None, position: int = None, bitrate: int = None,
            user_limit: int = None,
            permissions_overwrites: List[Overwrite] = None,
            rtc_region: str = None, video_quality_mod: int = None
    ) -> VoiceChannel:
        ...

    async def edit(self, **kwargs):
        """Edit a text channel with the given keyword arguments.

        Parameters
        ----------
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        return await super().edit(**kwargs)


class CategoryChannel(Channel):
    """A subclass of ``Channel`` for categories channels
    with all the same attributes.
    """
    pass


class NewsChannel(Channel):
    """A subclass of ``Channel`` for news channels with all the same attributes."""

    @overload
    async def edit(
            self, name: str = None, type: ChannelType = None,
            position: int = None, topic: str = None, nsfw: bool = None,
            permissions_overwrites: List[Overwrite] = None,
            parent_id: Snowflake = None,
            default_auto_archive_duration: int = None
    ) -> Union[TextChannel, NewsChannel]:
        ...

    async def edit(self, **kwargs):
        """Edit a text channel with the given keyword arguments.

        Parameters
        ----------
        \\*\\*kwargs :
            The keyword arguments to edit the channel with.

        Returns
        -------
        :class:`~pincer.objects.guild.channel.Channel`
            The updated channel object.
        """
        return await super().edit(**kwargs)


@dataclass
class ChannelMention(APIObject):
    """Represents a Discord Channel Mention object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the channel
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the guild containing the channel
    type: :class:`~pincer.objects.guild.channel.ChannelType`
        The type of channel
    name: :class:`str`
        The name of the channel
    """
    id: Snowflake
    guild_id: Snowflake
    type: ChannelType
    name: str


# noinspection PyTypeChecker
_channel_type_map: Dict[ChannelType, Channel] = {
    ChannelType.GUILD_TEXT: TextChannel,
    ChannelType.GUILD_VOICE: VoiceChannel,
    ChannelType.GUILD_CATEGORY: CategoryChannel,
    ChannelType.GUILD_NEWS: NewsChannel
}



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

import logging
from inspect import isasyncgenfunction, getfullargspec
from typing import Dict, Any
from typing import TYPE_CHECKING

from ..commands import ChatCommandHandler
from ..core.dispatch import GatewayDispatch
from ..objects import Interaction, MessageContext
from ..utils import MISSING, should_pass_cls, Coro, should_pass_ctx
from ..utils import get_index
from ..utils.conversion import construct_client_dict
from ..utils.signature import get_params, get_signature_and_params

if TYPE_CHECKING:
    from typing import List, Tuple


_log = logging.getLogger(__name__)


async def interaction_response_handler(
        self,
        command: Coro,
        context: MessageContext,
        interaction: Interaction,
        kwargs: Dict[str, Any]
):
    """|coro|

    Handle any coroutine as a command.

    Parameters
    ----------
    command : :data:`~pincer.utils.types.Coro`
        The coroutine which will be seen as a command.
    context : :class:`~pincer.objects.message.context.MessageContext`
        The context of the command.
    interaction : :class:`~pincer.objects.app.interactions.Interaction`
        The interaction which is linked to the command.
    \\*\\*kwargs :
        The arguments to be passed to the command.
    """
    if should_pass_cls(command):
        cls_keyword = getfullargspec(command).args[0]
        kwargs[cls_keyword] = ChatCommandHandler.managers[command.__module__]

    sig, params = get_signature_and_params(command)
    if should_pass_ctx(sig, params):
        kwargs[params[0]] = context

    if isasyncgenfunction(command):
        message = command(**kwargs)

        async for msg in message:
            if interaction.has_replied:
                await interaction.followup(msg)
            else:
                await interaction.reply(msg)
    else:
        message = await command(**kwargs)
        if not interaction.has_replied:
            await interaction.reply(message)


async def interaction_handler(
        self,
        interaction: Interaction,
        context: MessageContext,
        command: Coro
):
    """|coro|

    Processes an interaction.

    Parameters
    ----------
    interaction : :class:`~pincer.objects.app.interactions.Interaction`
        The interaction which is linked to the command.
    context : :class:`~pincer.objects.message.context.MessageContext`
        The context of the command.
    command : :data:`~pincer.utils.types.Coro`
        The coroutine which will be seen as a command.
    """
    self.throttler.handle(context)

    defaults = {param: None for param in get_params(command)}
    params = {}

    if interaction.data.options is not MISSING:
        params = {
            opt.name: opt.value for opt in interaction.data.options
        }

    kwargs = {**defaults, **params}

    await interaction_response_handler(
        self, command, context, interaction, kwargs
    )


async def interaction_create_middleware(
    self,
    payload: GatewayDispatch
) -> Tuple[str, List[Interaction]]:
    """Middleware for ``on_interaction``, which handles command
    execution.

    Parameters
    ----------
    payload : :class:`~pincer.core.dispatch.GatewayDispatch`
        The data received from the interaction event.


    Raises
    ------
    e
        Generic try except on ``await interaction_handler`` and
        ``if 0 < len(params) < 3``

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.app.interactions.Interaction`]]
        ``on_interaction_create`` and an ``Interaction``
    """  # noqa: E501
    interaction: Interaction = Interaction.from_dict(
        construct_client_dict(self, payload.data)
    )
    await interaction.build()
    command = ChatCommandHandler.register.get(interaction.data.name)

    if command:
        context = interaction.convert_to_message_context(command)

        try:
            await interaction_handler(self, interaction, context,
                                      command.call)
        except Exception as e:
            if coro := get_index(self.get_event_coro("on_command_error"), 0):
                params = get_signature_and_params(coro)[1]

                # Check if a context or error var has been passed.
                if 0 < len(params) < 3:
                    await interaction_response_handler(
                        self,
                        coro,
                        context,
                        interaction,
                        # Always take the error parameter its name.
                        {params[(len(params) - 1) or 0]: e}
                    )
                else:
                    raise e
            else:
                raise e

    return "on_interaction_create", [interaction]


def export() -> Coro:
    return interaction_create_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from ...utils.snowflake import Snowflake


class PrivacyLevel(IntEnum):
    """Represents the level of publicity of a stage.

    Attributes
    ----------
    PUBLIC:
        The stage is public.
    GUILD_ONLY:
        The stage of for guild members only.
    """
    PUBLIC = 1
    GUILD_ONLY = 2


@dataclass
class StageInstance(APIObject):
    """Represents a Stage Instance object

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of this Stage instance
    guild_id: :class:`~pincer.utils.snowflake.Snowflake`
        Guild id of the associated Stage channel
    channel_id: :class:`~pincer.utils.snowflake.Snowflake`
        Id of the associated Stage channel
    topic: :class:`str`
        Topic of the Stage instance (1-120 characters)
    privacy_level: :class:`~pincer.objects.guild.stage.PrivacyLevel`
        Privacy level of the Stage instance
    discoverable: :class:`bool`
        Is Stage Discovery enabled
    """
    id: Snowflake
    guild_id: Snowflake
    channel_id: Snowflake
    topic: str
    privacy_level: PrivacyLevel
    discoverable: bool



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from enum import IntEnum


class AppCommandType(IntEnum):
    """
    Defines the different types of application commands.

    Attributes
    ----------
    CHAT_INPUT:
        Slash commands; a text-based command that shows up when a user
        types /

    USER:
        A UI-based command that shows up when you right click or tap on
        a user

    MESSAGE:
        A UI-based command that shows up when you right click or tap on
        a message
    """
    CHAT_INPUT = 1
    USER = 2
    MESSAGE = 3


class AppCommandOptionType(IntEnum):
    """
    Represents a parameter type.

    Attributes
    ----------
    SUB_COMMAND:
        The parameter will be a subcommand.

    SUB_COMMAND_GROUP:
        The parameter will be a group of subcommands.

    STRING:
        The parameter will be a string.

    INTEGER:
        The parameter will be an integer/number. (-2^53 and 2^53)

    BOOLEAN:
        The parameter will be a boolean.

    USER:
        The parameter will be a Discord user object.

    CHANNEL:
        The parameter will be a Discord channel object.

    ROLE:
        The parameter will be a Discord role object.

    MENTIONABLE:
        The parameter will be mentionable.

    NUMBER:
        The parameter will be a float. (-2^53 and 2^53)
    """
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4  # 54-bit
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7
    ROLE = 8
    MENTIONABLE = 9
    NUMBER = 10  # 54-bit



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when the thread member object for the current user is updated"""

from ..core.dispatch import GatewayDispatch
from ..objects import ThreadMember
from ..utils import Timestamp
from ..utils.conversion import construct_client_dict


async def thread_member_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_thread_member_update`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the thread member update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.thread.ThreadMember`]]
        ``on_thread_member_update`` and an ``ThreadMember``
    """

    return "on_thread_member_update", [
        ThreadMember.from_dict(construct_client_dict(
            self,
            {
                "join_timestamp": Timestamp(payload.data.pop("join_timestamp")),
                **payload.data
            }
        ))
    ]


def export():
    return thread_member_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake


@dataclass
class RoleTags(APIObject):
    """Special tags/flags which have been defined for a role.

    Attributes
    ----------
    bot_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the bot this role belongs to.
        (the role got created by adding the bot with this id)
    integration_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The id of the integration this role belongs to.
        (the role got created by adding an integration with this id)
    premium_subscriber: APINullable[:class:`bool`]
        Whether this is the guild's premium subscriber role or not.
    """
    bot_id: APINullable[Snowflake] = MISSING
    integration_id: APINullable[Snowflake] = MISSING
    premium_subscriber: APINullable[bool] = MISSING


@dataclass
class Role(APIObject):
    """
    Represents a Discord guild/server role.

    Attributes
    ----------
    color: :class:`int`
        Integer representation of hexadecimal color code
    hoist: :class:`bool`
        If this role is pinned in the user listing
    id: :class:`~pincer.utils.snowflake.Snowflake`
        Role id
    managed: :class:`bool`
        Whether this role is managed by an integration
    mentionable: :class:`bool`
        Whether this role is mentionable
    name: :class:`str`
        Role name
    permissions:
        Permission bit set
    position: :class:`int`
        Position of this role
    icon: APINullable[:class:`str`]
        The role's icon
    unicode_emoji: APINullable[:class:`str`]
        The unicode emoji for this role
    tags: :class:`~pincer.objects.guild.role.RoleTags`
        The tags this role has
    """
    color: int
    hoist: bool
    id: Snowflake
    managed: bool
    mentionable: bool
    name: str
    permissions: str
    position: int

    icon: APINullable[str] = MISSING
    unicode_emoji: APINullable[str] = MISSING
    tags: APINullable[RoleTags] = MISSING

    @property
    def mention(self) -> str:
        """:class:`str`\\: Returns the mention string for this role."""
        return f"<@&{self.id}>"

    # TODO: Implement Caching @Arthurdw
    @classmethod
    async def from_id(cls, client, guild_id: int, role_id: int) -> Role:
        roles: list = await client.http.get(f"/guilds/{guild_id}/roles")

        for role in roles:
            if int(role['id']) == role_id:
                return cls.from_dict(role)



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass, MISSING
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject

if TYPE_CHECKING:
    from typing import List

    from .user import VisibilityType
    from .integration import Integration
    from ...utils.types import APINullable


@dataclass
class Connection(APIObject):
    """The connection object that the user has attached.

    Attributes
    ----------
    id: :class:`str`
        Id of the connection account
    name: :class:`str`
        The username of the connection account
    type: :class:`str`
        The service of the connection (twitch, youtube)
    verified: :class:`bool`
        Whether the connection is verified
    friend_sync: :class:`bool`
        Whether friend sync is enabled for this connection
    show_activity: :class:`bool`
        Whether activities related to this connection
        will be shown in presence updates
    visibility: :class:`~pincer.objects.user.user.VisibilityType`
        If the connection is visible
    revoked: APINullable[:class:`bool`]
        Whether the connection is revoked
    integrations: APINullable[List[:class:`~pincer.objects.user.integration.Integration`]]
        An array of partial server integrations
    """
    # noqa: E501

    id: str
    name: str
    type: str
    verified: bool
    friend_sync: bool
    show_activity: bool
    visibility: VisibilityType

    revoked: APINullable[bool] = MISSING
    integrations: APINullable[List[Integration]] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when the client gains access to a channel"""

from __future__ import annotations

from typing import List

from ..core.dispatch import GatewayDispatch
from ..objects import Channel, ThreadMember
from ..objects.events.thread import ThreadListSyncEvent
from ..utils.conversion import construct_client_dict


async def thread_list_sync(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_thread_list_sync`` event.

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the thread list sync event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.events.thread.ThreadListSyncEvent`]]
        ``on_thread_list_sync`` and an ``ThreadListSyncEvent``
    """

    threads: List[Channel] = [
        Channel.from_dict(construct_client_dict(self, thread))
        for thread in payload.data.pop("threads")
    ]

    members: List[ThreadMember] = [
        ThreadMember.from_dict(construct_client_dict(self, member))
        for member in payload.data.pop("members")
    ]

    return "on_thread_list_sync", [
        ThreadListSyncEvent.from_dict(
            {"threads": threads, "members": members, **payload.data}
        )
    ]


def export():
    return thread_list_sync



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:

    from typing import Optional, Union
    from .user_message import UserMessage
    from ..app import ClientCommandStructure, Interaction
    from ..app.interaction_flags import InteractionFlags
    from ..guild.member import GuildMember
    from ..user.user import User
    from ...utils.convert_message import MessageConvertable
    from ...utils.snowflake import Snowflake


@dataclass
class MessageContext:
    """Represents the context of a message interaction.

    Attributes
    ----------
    id: :class:`~pincer.utils.snowflake.Snowflake`
        The ID of the interaction.
    author: Union[:class:`~pincer.objects.guild.member.GuildMember`, :class:`~pincer.objects.user.user.User`]

        The user whom invoked the interaction.
    command: :class:`~pincer.objects.app.command.ClientCommandStructure`
        The local command object for the command to whom this context
        belongs.

    guild_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The ID of the guild the interaction was invoked in.
        Can be None if it wasn't invoked in a guild.
    channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The ID of the channel the interaction was invoked in.
        Can be None if it wasn't invoked in a channel.
    """
    # noqa: E501
    author: Union[GuildMember, User]
    command: ClientCommandStructure
    interaction: Interaction

    guild_id: Optional[Snowflake] = None
    channel_id: Optional[Snowflake] = None

    async def ack(self, flags: InteractionFlags):
        """|coro|

        Alias for :func:`~pincer.objects.app.interactions.Interaction.ack`.

        Parameters
        ----------
        flags :class:`~pincer.objects.app.interaction_flags.InteractionFlags`
            The flags which must be applied to the reply.
        """
        await self.interaction.ack(flags)

    async def reply(self, message: MessageConvertable):
        """|coro|

        Alias for :func:`~pincer.objects.app.interactions.Interaction.reply`.

        Parameters
        ----------
        message :class:`~pincer.utils.convert_message.MessageConvertable`
            The response message!
        """
        await self.interaction.reply(message)

    async def followup(self, message: MessageConvertable) -> UserMessage:
        """|coro|

        Alias for :func:`~pincer.objects.app.interactions.Interaction.followup`.

        Parameters
        ----------
        message :class:`~pincer.utils.convert_message.MessageConvertable`
            The message to sent.
        """
        return await self.interaction.followup(message)

    async def send(self, message: MessageConvertable) -> UserMessage:
        """|coro|

        Send a response for an interaction.
        This object returns the sent object and may be used several
        times after each other. (first one will always be the main
        interaction response)

        Uses
        ----
        :func:`~pincer.objects.message.context.MessageContext.reply`
            Method gets called for initial send.
        :func:`~pincer.objects.app.interactions.Interaction.response`
            Method gets called for initial send to get response.
        :func:`~pincer.objects.message.context.MessageContext.followup`
            Method gets called for second message and onwards.

         Returns
         -------
         :class:`~pincer.objects.message.user_message.UserMessage`
            The message that was sent.
        """
        if self.interaction.has_replied:
            return await self.followup(message)

        await self.reply(message)
        return await self.interaction.response()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.
from enum import Enum, auto


class ThrottleScope(Enum):
    """On what the cooldown should be set/on what should the cooldown be
    set.

    Attributes
    ----------
    GUILD:
        The cooldown is per guild.
    CHANNEL:
        The cooldown is per channel.
    USER:
        The cooldown is per user.
    GLOBAL:
        The cooldown is global.
    """
    GUILD = auto()
    CHANNEL = auto()
    USER = auto()
    GLOBAL = auto()



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from enum import IntEnum


class InteractionFlags(IntEnum):
    """

    Attributes
    ----------
    EPHEMERAL:
        only the user receiving the message can see it
    """
    EPHEMERAL = 1 << 6



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ...utils.api_object import APIObject


@dataclass
class VoiceRegion(APIObject):
    """Represents a Discord Voice Region object

    Attributes
    ----------
    id: :class:`str`
        Unique ID for the region
    name: :class:`str`
        Name of the region
    vip: :class:`bool`
        True if this is a vip-only server
    optimal: :class:`bool`
        True for a single server
        that is closest to the current user's client
    deprecated: :class:`bool`
        Whether this is a deprecated voice region
        (avoid switching to these)
    custom: :class:`bool`
        Whether this is a custom voice region
        (used for events/etc)
    """
    id: str
    name: str
    vip: bool
    optimal: bool
    deprecated: bool
    custom: bool



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from dataclasses import dataclass

from ...utils import APIObject


@dataclass
class ActivityJoinEvent(APIObject):
    """
    Sent when the user clicks a Rich Presence join invite in chat to join a game.

    Attributes
    ----------
    secret : :class:`str`
        the join_secret for the given invite
    """

    secret: str


@dataclass
class ActivitySpectateEvent(APIObject):
    """
    Sent when the user clicks a Rich Presence spectate invite in chat to spectate a game.

    Attributes
    ----------
    secret: :class:`str`
        the spectate_secret for the given invite
    """

    secret: str



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

from ...utils.api_object import APIObject
from ...utils.types import MISSING

if TYPE_CHECKING:
    from typing import Optional

    from ..guild.member import GuildMember
    from ...utils.types import APINullable
    from ...utils.snowflake import Snowflake
    from ...utils.timestamp import Timestamp


@dataclass
class VoiceState(APIObject):
    """Used to represent a user's voice connection status

    Attributes
    ----------
    channel_id: Optional[:class:`~pincer.utils.snowflake.Snowflake`]
        The channel id this user is connected to
    user_id: :class:`~pincer.utils.snowflake.Snowflake`
        The user id this voice state is for
    session_id: :class:`str`
        The session id for this voice state
    deaf: :class:`bool`
        Whether this user is deafened by the server
    mute: :class:`bool`
        Whether this user is muted by the server
    self_deaf: :class:`bool`
        Whether this user is locally deafened
    self_mute: :class:`bool`
        Whether this user is locally muted
    self_video: :class:`bool`
        Whether this user's camera is enabled
    suppress: :class:`bool`
        Whether this user is muted by the current user
    request_to_speak_timestamp: Optional[:class:`~pincer.utils.timestamp.Timestamp`]
        The time at which the user requested to speak
    guild_id: APINullable[:class:`~pincer.utils.snowflake.Snowflake`]
        The guild id this voice state is for
    member: APINullable[:class:`~pincer.objects.guild.member.GuildMember`]
        The guild member this voice state is for
    self_stream: APINullable[:class:`bool`]
        Whether this user is streaming using "Go Live"
    """
    # noqa: E501

    channel_id: Optional[Snowflake]
    user_id: Snowflake
    session_id: str
    deaf: bool
    mute: bool
    self_deaf: bool
    self_mute: bool
    self_video: bool
    suppress: bool
    request_to_speak_timestamp: Optional[Timestamp]

    guild_id: APINullable[Snowflake] = MISSING
    member: APINullable[GuildMember] = MISSING
    self_stream: APINullable[bool] = MISSING



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when a message in pinned/unpinned"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.channel import ChannelPinsUpdateEvent


async def channel_pins_update_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_channel_pins_update``,

    Parameters
    ----------
    payload : :class:`pincer.core.dispatch.GatewayDispatch`
        The data received from the channel pins update event.

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.objects.guild.channel.Channel`]]
        ``on_channel_pins_update`` and a ``Channel``
    """

    return "on_channel_pins_update", [
        ChannelPinsUpdateEvent.from_dict(payload.data)
    ]


def export():
    return channel_pins_update_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

"""sent when multiple messages are deleted at once"""

from ..core.dispatch import GatewayDispatch
from ..objects.events.message import MessageDeleteBulkEvent
from ..utils.conversion import construct_client_dict
from ..utils.types import Coro


async def message_delete_bulk_middleware(self, payload: GatewayDispatch):
    """|coro|

    Middleware for ``on_message_delete_bulk``,

    Parameters
    ----------
    payload : :class:`GatewayDispatch`
        The data received from the message delete bulk event

    Returns
    -------
    Tuple[:class:`str`, List[:class:`~pincer.events.message.MessageDeleteBulkEvent`]]
        ``on_message_delete_bulk`` and an ``MessageDeleteBulkEvent``
    """
    return "on_message_delete_bulk", [
        MessageDeleteBulkEvent.from_dict(construct_client_dict(self, payload.data))
    ]


def export() -> Coro:
    return message_delete_bulk_middleware



# Copyright Pincer 2021-Present
# Full MIT License can be found in `LICENSE` at the project root.

from .connection import Connection
from .integration import (
    IntegrationExpireBehavior, IntegrationApplication,
    Integration, IntegrationAccount
)
from .user import User, PremiumTypes, VisibilityType
from .voice_state import VoiceState


__all__ = (
    "Connection", "Integration", "IntegrationAccount",
    "IntegrationApplication", "IntegrationExpireBehavior", "PremiumTypes",
    "User", "VisibilityType", "VoiceState"
)
